# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
from pathlib import Path

from compiler_config.config import CompilerConfig, Tket, TketOptimizations

from qat.core.metrics_base import MetricsManager
from qat.core.pass_base import PassManager
from qat.core.result_base import ResultManager
from qat.frontend.base import BaseFrontend
from qat.frontend.parsers.qir import QIRParser as PydQIRParser
from qat.integrations.tket import TketBuilder, TketToQatIRConverter, run_tket_optimizations
from qat.ir.builder_factory import BuilderFactory
from qat.model.hardware_model import PhysicalHardwareModel as PydHardwareModel
from qat.purr.compiler.builders import InstructionBuilder
from qat.purr.compiler.hardware_models import QuantumHardwareModel
from qat.purr.integrations.qir import QIRParser
from qat.purr.integrations.tket import run_tket_optimizations_qir

string_regex = r"@__quantum__qis"


def is_qir_path(src: str) -> bool:
    """Determines if the string corresponds to a path, and has a `.ll` or `.rb`
    extension."""

    return src.endswith(".ll") or src.endswith(".bc")


def is_qir_str(src: str) -> bool:
    """Determines if the string contains QIR semantics."""

    if isinstance(src, str) and string_regex in src:
        return True
    return False


def load_qir_file(path: str) -> str | bytes:
    """Loads a QIR file from a given file path.

    :param path: The file path.
    :raises ValueError: Raises an error if the the file has an invalid extension.
    :returns: The QIR program as a string or bytes.
    """

    path = Path(path)
    if path.suffix == ".ll":
        with path.open() as file:
            src = file.read()
    elif path.suffix == ".bc":
        with path.open("rb") as file:
            src = file.read()
    else:
        raise ValueError("String expected to end in `.ll` or `.bc`.")
    return src


class QIRFrontend(BaseFrontend):
    """A frontend for handling QIR programs.

    Handles the parsing of QIR programs, and optionally implements a pipeline for validation
    and transformation.
    """

    def __init__(
        self,
        model: QuantumHardwareModel | PydHardwareModel,
        pipeline: None | PassManager = None,
    ):
        """
        :param model: The hardware model can be required for the pipeline and is used in
            parsing.
        :param pipeline: A pipeline for validation and optimising QASM, defaults to a
            predefined pipeline that optimizes the QASM file.
        """
        super().__init__(model)
        self.pipeline = pipeline if pipeline is not None else self.build_pass_pipeline()
        self._pyd_model = isinstance(model, PydHardwareModel)

    @staticmethod
    def build_pass_pipeline() -> PassManager:
        """Creates an empty pipeline for QIR compilation.

        :return: The pipeline as a pass manager.
        """

        return PassManager()

    def _init_parser(
        self, config: CompilerConfig | None = None
    ) -> QIRParser | PydQIRParser:
        results_format = (
            config.results_format.format if config and config.results_format else None
        )
        if not self._pyd_model:
            return QIRParser(self.model, results_format=results_format)
        else:
            return PydQIRParser(results_format)

    def check_and_return_source(self, src: str | bytes) -> bool | str | bytes:
        """Checks that the source program (or file path) can be interpreted as a QIR file
        by checking against the language.

        :param src: The QIR program, or the file path to the program.
        :returns: If the program is determined to not be valid, False is returned.
            Otherwise, the program is returned (and loaded if required).
        """
        if not isinstance(src, (str, bytes)):
            return False

        if isinstance(src, str):
            if src.endswith(".ll") or src.endswith(".bc"):
                src = load_qir_file(src)

        if isinstance(src, bytes):
            if src.startswith(b"BC"):
                return src
            else:
                return False

        if is_qir_str(src):
            return src

        return False

    def _check_source(self, src: str) -> str:
        src = self.check_and_return_source(src)
        if not src:
            raise ValueError(
                "Source program is not a QIR program, or a path to a QIR program."
            )
        return src

    def _get_builder(
        self, compiler_config: CompilerConfig, src: str
    ) -> InstructionBuilder | None:
        builder = BuilderFactory.create_builder(self.model)
        parser = self._init_parser(compiler_config)
        optimizations = compiler_config.optimizations
        if not self._pyd_model:
            if (
                isinstance(optimizations, Tket)
                and optimizations.tket_optimizations != TketOptimizations.Empty
            ):
                builder = run_tket_optimizations_qir(
                    src,
                    optimizations.tket_optimizations,
                    self.model,
                    compiler_config.results_format.format,
                )
            else:
                builder = parser.parse(src)

            builder = (
                self.model.create_builder()
                .repeat(
                    compiler_config.repeats,
                    repetition_period=compiler_config.repetition_period,
                    passive_reset_time=compiler_config.passive_reset_time,
                )
                .add(builder)
            )
        else:
            builder.repeat(compiler_config.repeats)
            if (
                isinstance(optimizations, Tket)
                and optimizations.tket_optimizations != TketOptimizations.Empty
            ):
                tket_builder = TketBuilder(self.model)
                tket_builder = parser.parse(tket_builder, src)
                tket_builder.circuit = run_tket_optimizations(
                    tket_builder.circuit,
                    optimizations.tket_optimizations,
                    self.model,
                    return_as_qasm_str=False,
                )
                builder = TketToQatIRConverter().convert(builder, tket_builder)
            else:
                builder = parser.parse(builder, src)
        return builder

    def emit(
        self,
        src: str,
        res_mgr: ResultManager | None = None,
        met_mgr: MetricsManager | None = None,
        compiler_config: CompilerConfig | None = None,
        **kwargs,
    ) -> InstructionBuilder:
        """Compiles the source QIR program into QAT's intermediate representation.

        :param src: The source program, or path to the source program.
        :param res_mgr: The result manager to store results from passes within the pipeline,
            defaults to an empty :class:`ResultManager`.
        :param met_mgr: The metrics manager to store metrics, such as the optimized QIR
            circuit, defaults to an empty :class:`MetricsManager`.
        :param compiler_config: The compiler config is used in both the pipeline and for
            parsing.
        :raises ValueError: An error is thrown if the the program is not detected to be a
            QIR program.
        :returns: The program as an :class:`InstructionBuilder`.
        """
        res_mgr, met_mgr, compiler_config = self._check_metrics_and_config(
            res_mgr, met_mgr, compiler_config
        )

        src = self._check_source(src)
        src = self.pipeline.run(
            src, res_mgr, met_mgr, compiler_config=compiler_config, **kwargs
        )
        builder = self._get_builder(compiler_config, src)

        return builder



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024-2025 Oxford Quantum Circuits Ltd

from abc import ABC, abstractmethod

from qblox_instruments import Module, Sequencer

from qat.backend.qblox.config.specification import (
    MixerConfig,
    ModuleConfig,
    OffsetConfig,
    SequencerConfig,
)
from qat.purr.utils.logger import get_default_logger

log = get_default_logger()


class QbloxConfigHelper(ABC):
    def __init__(
        self,
        module_config: ModuleConfig = None,
        sequencer_config: SequencerConfig = None,
    ):
        self.sequencer_config = sequencer_config or SequencerConfig()
        self.module_config = module_config or SequencerConfig()

    def configure(self, module: Module, sequencer: Sequencer):
        self.configure_module(module)
        self.configure_sequencer(sequencer)

    @abstractmethod
    def configure_module(self, module: Module):
        pass

    @abstractmethod
    def configure_sequencer(self, sequencer: Sequencer):
        pass

    def configure_connection(self, sequencer: Sequencer):
        if self.sequencer_config.connection.bulk_value:
            sequencer.connect_sequencer(*self.sequencer_config.connection.bulk_value)

    def configure_nco(self, sequencer: Sequencer):
        if self.sequencer_config.nco.freq:
            sequencer.nco_freq(self.sequencer_config.nco.freq)
        if self.sequencer_config.nco.prop_delay_comp_en:
            sequencer.nco_prop_delay_comp_en(self.sequencer_config.nco.prop_delay_comp_en)
        if self.sequencer_config.nco.prop_delay_comp:
            sequencer.nco_prop_delay_comp(self.sequencer_config.nco.prop_delay_comp)
        if self.sequencer_config.nco.phase_offs:
            sequencer.nco_phase_offs(self.sequencer_config.nco.phase_offs)

    def configure_awg(self, sequencer: Sequencer):
        if self.sequencer_config.awg.mod_en:
            sequencer.mod_en_awg(self.sequencer_config.awg.mod_en)
        if self.sequencer_config.awg.gain_path0:
            sequencer.gain_awg_path0(self.sequencer_config.awg.gain_path0)
        if self.sequencer_config.awg.gain_path1:
            sequencer.gain_awg_path1(self.sequencer_config.awg.gain_path1)
        if self.sequencer_config.awg.offset_path0:
            sequencer.offset_awg_path0(self.sequencer_config.awg.offset_path0)
        if self.sequencer_config.awg.offset_path1:
            sequencer.offset_awg_path1(self.sequencer_config.awg.offset_path1)

        if self.sequencer_config.awg.cont_mode_en_path0:
            sequencer.cont_mode_en_awg_path0(self.sequencer_config.awg.cont_mode_en_path0)
        if self.sequencer_config.awg.cont_mode_en_path1:
            sequencer.cont_mode_en_awg_path1(self.sequencer_config.awg.cont_mode_en_path1)
        if self.sequencer_config.awg.cont_mode_waveform_idx_path0:
            sequencer.cont_mode_waveform_idx_awg_path0(
                self.sequencer_config.awg.cont_mode_waveform_idx_path0
            )
        if self.sequencer_config.awg.cont_mode_waveform_idx_path1:
            sequencer.cont_mode_waveform_idx_awg_path1(
                self.sequencer_config.awg.cont_mode_waveform_idx_path1
            )

        if self.sequencer_config.awg.cont_mode_en_path0:
            sequencer.cont_mode_en_awg_path0(self.sequencer_config.awg.cont_mode_en_path0)
        if self.sequencer_config.awg.cont_mode_en_path1:
            sequencer.cont_mode_en_awg_path1(self.sequencer_config.awg.cont_mode_en_path1)
        if self.sequencer_config.awg.cont_mode_waveform_idx_path0:
            sequencer.cont_mode_waveform_idx_awg_path0(
                self.sequencer_config.awg.cont_mode_waveform_idx_path0
            )
        if self.sequencer_config.awg.cont_mode_waveform_idx_path1:
            sequencer.cont_mode_waveform_idx_awg_path1(
                self.sequencer_config.awg.cont_mode_waveform_idx_path1
            )

    def configure_mixer(self, sequencer: Sequencer):
        if self.sequencer_config.mixer.phase_offset:
            sequencer.mixer_corr_phase_offset_degree(
                self.sequencer_config.mixer.phase_offset
            )
        if self.sequencer_config.mixer.gain_ratio:
            sequencer.mixer_corr_gain_ratio(self.sequencer_config.mixer.gain_ratio)

    def calibrate_mixer(
        self, module: Module, sequencer: Sequencer = None, connection: str = None
    ):
        """
        - Compensates for the LO leakage for input(s)/output(s) depending on module
        - Suppresses undesired sideband on sequencer

        :param module: A Module instance representing a QCM-RF or a QRM-RF card.
        :param sequencer: A Sequencer instance representing the sequencer.
            When None, sideband cal will run on all sequencers belonging to module `module`
        :param connection: An optional string indicating muxing of Sequencer to analog channel
            Defaults to "out0"
        """

        sequencers = [sequencer] if sequencer else module.sequencers

        if connection:
            log.info(
                f"Provided argument `connection` is {connection}. Will overwrite sequencer's connections"
            )
            module.disconnect_outputs()
            for sequencer in sequencers:
                sequencer.connect_sequencer(connection)
        else:
            log.info(
                "Argument `connection` not provided. Assuming `connection` already specified"
            )

        log.info(f"Calibrating LO leakage on module {module}")
        offset_config = self.calibrate_lo_leakage(module)
        mixer_configs = {}
        for sequencer in sequencers:
            mixer_configs[sequencer.seq_idx] = self.calibrate_sideband(
                sequencer, connection
            )

        if connection:
            log.info(
                f"Mixer calibration finished. Resetting sequencer connections for module {module.slot_idx}"
            )
            module.disconnect_outputs()

        return offset_config, mixer_configs

    @abstractmethod
    def calibrate_lo_leakage(self, module: Module):
        pass

    def calibrate_sideband(self, sequencer: Sequencer, connection: str = None):
        log.info(f"Calibrating sidebands on sequencer {sequencer}")
        sequencer.sideband_cal()
        sequencer.arm_sequencer()
        sequencer.start_sequencer()

        return MixerConfig(
            phase_offset=sequencer.mixer_corr_phase_offset_degree(),
            gain_ratio=sequencer.mixer_corr_gain_ratio(),
        )


class QcmConfigHelper(QbloxConfigHelper):
    pass


class QcmRfConfigHelper(QcmConfigHelper):
    def configure_module(self, module: Module):
        self.configure_lo(module)
        self.configure_attenuation(module)
        self.configure_offset(module)

    def configure_sequencer(self, sequencer: Sequencer):
        self.configure_connection(sequencer)
        self.configure_nco(sequencer)
        self.configure_awg(sequencer)
        self.configure_mixer(sequencer)

    def configure_lo(self, module: Module):
        if self.module_config.lo.out0_en:
            module.out0_lo_en(self.module_config.lo.out0_en)  # Switch the LO 0 on
        if self.module_config.lo.out0_freq:
            module.out0_lo_freq(self.module_config.lo.out0_freq)
        if self.module_config.lo.out1_en:
            module.out1_lo_en(self.module_config.lo.out1_en)  # Switch the LO 1 on
        if self.module_config.lo.out1_freq:
            module.out1_lo_freq(self.module_config.lo.out1_freq)

    def configure_attenuation(self, module: Module):
        if self.module_config.attenuation.out0:
            module.out0_att(self.module_config.attenuation.out0)
        if self.module_config.attenuation.out1:
            module.out1_att(self.module_config.attenuation.out1)

    def configure_offset(self, module: Module):
        if self.module_config.offset.out0_path0:
            module.out0_offset_path0(self.module_config.offset.out0_path0)
        if self.module_config.offset.out0_path1:
            module.out0_offset_path1(self.module_config.offset.out0_path1)
        if self.module_config.offset.out1_path0:
            module.out1_offset_path0(self.module_config.offset.out1_path0)
        if self.module_config.offset.out1_path1:
            module.out1_offset_path1(self.module_config.offset.out1_path1)

    def calibrate_lo_leakage(self, module: Module):
        module.out0_lo_cal()
        module.out1_lo_cal()

        return OffsetConfig(
            out0_path0=module.out0_offset_path0(),
            out0_path1=module.out0_offset_path1(),
            out1_path0=module.out1_offset_path0(),
            out1_path1=module.out1_offset_path1(),
        )


class QrmConfigHelper(QbloxConfigHelper):
    def configure_scope_acq(self, module: Module):
        scope_acq = self.module_config.scope_acq
        if scope_acq.sequencer_select:
            module.scope_acq_sequencer_select(scope_acq.sequencer_select)
        if scope_acq.trigger_mode_path0:
            module.scope_acq_trigger_mode_path0(scope_acq.trigger_mode_path0)
        if scope_acq.avg_mode_en_path0:
            module.scope_acq_avg_mode_en_path0(scope_acq.avg_mode_en_path0)
        if scope_acq.trigger_mode_path1:
            module.scope_acq_trigger_mode_path1(scope_acq.trigger_mode_path1)
        if scope_acq.avg_mode_en_path1:
            module.scope_acq_avg_mode_en_path1(scope_acq.avg_mode_en_path1)


class QrmRfConfigHelper(QrmConfigHelper):
    def configure_module(self, module: Module):
        self.configure_lo(module)
        self.configure_attenuation(module)
        self.configure_offset(module)
        self.configure_scope_acq(module)

    def configure_sequencer(self, sequencer: Sequencer):
        self.configure_connection(sequencer)
        self.configure_nco(sequencer)
        self.configure_awg(sequencer)
        self.configure_mixer(sequencer)
        self.configure_acq(sequencer)

    def configure_lo(self, module: Module):
        if self.module_config.lo.out0_in0_en:
            module.out0_in0_lo_en(self.module_config.lo.out0_in0_en)  # Switch the LO on
        if self.module_config.lo.out0_in0_freq:
            module.out0_in0_lo_freq(self.module_config.lo.out0_in0_freq)

    def configure_attenuation(self, module: Module):
        if self.module_config.attenuation.out0:
            module.out0_att(self.module_config.attenuation.out0)
        if self.module_config.attenuation.in0:
            module.in0_att(self.module_config.attenuation.in0)

    def configure_offset(self, module: Module):
        if self.module_config.offset.out0_path0:
            module.out0_offset_path0(self.module_config.offset.out0_path0)
        if self.module_config.offset.out0_path1:
            module.out0_offset_path1(self.module_config.offset.out0_path1)
        if self.module_config.offset.in0_path0:
            module.in0_offset_path0(self.module_config.offset.in0_path0)
        if self.module_config.offset.in0_path1:
            module.in0_offset_path1(self.module_config.offset.in0_path1)

    def configure_acq(self, sequencer: Sequencer):
        # Square weight integration
        if self.sequencer_config.demod_en_acq:
            sequencer.demod_en_acq(self.sequencer_config.demod_en_acq)
        if self.sequencer_config.square_weight_acq.integration_length:
            sequencer.integration_length_acq(
                self.sequencer_config.square_weight_acq.integration_length
            )
        if self.sequencer_config.thresholded_acq.rotation:
            sequencer.thresholded_acq_rotation(
                self.sequencer_config.thresholded_acq.rotation
            )
        if self.sequencer_config.thresholded_acq.threshold:
            sequencer.thresholded_acq_threshold(
                self.sequencer_config.thresholded_acq.threshold
            )
        if self.sequencer_config.thresholded_acq.trigger_en:
            sequencer.thresholded_acq_trigger_en(
                self.sequencer_config.thresholded_acq.trigger_en
            )
        if self.sequencer_config.thresholded_acq.trigger_address:
            sequencer.thresholded_acq_trigger_address(
                self.sequencer_config.thresholded_acq.trigger_address
            )
        if self.sequencer_config.thresholded_acq.trigger_invert:
            sequencer.thresholded_acq_trigger_invert(
                self.sequencer_config.thresholded_acq.trigger_invert
            )

    def calibrate_lo_leakage(self, module: Module):
        module.out0_in0_lo_cal()

        return OffsetConfig(
            out0_path0=module.out0_offset_path0(),
            out0_path1=module.out0_offset_path1(),
            in0_path0=module.in0_offset_path0(),
            in0_path1=module.in0_offset_path1(),
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
from qat.backend.waveform.codegen import PydWaveformBackend
from qat.frontend import AutoFrontendWithFlattenedIR
from qat.middleend import PydDefaultMiddleend
from qat.model.hardware_model import PhysicalHardwareModel
from qat.model.target_data import TargetData
from qat.pipelines.pipeline import CompilePipeline
from qat.pipelines.updateable import PipelineConfig, UpdateablePipeline
from qat.purr.utils.logger import get_default_logger

log = get_default_logger()


class WaveformCompilePipeline(UpdateablePipeline):
    """A pipeline that compiles programs using the :class:`PydWaveformBackend`.

    .. warning::

        This pipeline is for compilation purposes only and does not execute programs.
        Please select an appropriate execution pipeline if you wish to execute compiled
        programs.
    """

    @staticmethod
    def _build_pipeline(
        config: PipelineConfig,
        model: PhysicalHardwareModel,
        target_data: TargetData | None,
        engine: None = None,
    ) -> CompilePipeline:
        """Constructs a pipeline equipped with the :class:`PydWaveformBackend`
        and :class:`EchoEngine`."""

        if engine is not None:
            log.warning(
                "The compilation pipeline does not require an engine. It will be ignored."
            )

        target_data = target_data if target_data is not None else TargetData.default()
        return CompilePipeline(
            model=model,
            target_data=target_data,
            frontend=AutoFrontendWithFlattenedIR.default_for_pydantic(model),
            middleend=PydDefaultMiddleend(model, target_data),
            backend=PydWaveformBackend(model, target_data),
            name=config.name,
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
from enum import Flag, auto


class ConnectionMode(Flag):
    """The connection mode can be used to specify how a Runtime should handle connections.

    The method for connection handling will depend in the context in which the Runtime is
    used. For example, if a user only requires to send jobs sporadically, then a connection
    that is created before execution and ended after execution might be desired. On the
    contrary, if there is a stream of frequent jobs, the connection likely wants to be held
    indefinitely.

    There are different flags that can be specified:

    #. `ConnectionMode.CONNECT_BEFORE_EXECUTE`: Specifies that the Runtime should attempt
        to connect the engine each time execute is called.
    #. `ConnectionMode.DISCONNECT_AFTER_EXECUTE`: Specifies that the Runtime should attempt
        to disconnect the engine after each execute.
    #. `ConnectionMode.CONNECT_AT_BEGINNING`: Specifies that the Runtime should attempt to
        connect the engine as soon as its instantiated.
    #. `ConnectionMode.DISCONNECT_AT_END`: Specifies that the Runtime should attempt to
        disconnect the engine when the Runtime is deleted (or at the end of a context
        manager).

    Some predefined composite modes:

    #. `ConnectionMode.MANUAL`: Connection handling is done externally: the Runtime will
        make no attempt to manage it.
    #. `ConnectionMode.DEFAULT`: The connection will be created before each execution and
        disconnected immediately after the execution is complete.
    #. `ConnectionMode.ALWAYS`: A connection is created at instantiation of the Runtime,
        and is kept throughout the lifetime of the Runtime. It will not try to connect at
        execution.
    #. `ConnectionMode.ALWAYS_ON_EXECUTE`: A connection is attempted each time execution is
        called. However, the Runtime will not attempt to disconnect after execution is
        complete. This allows for external disconnection of the engine during e.g. idle
        periods, and for the connection to be restored when needed.

    .. warning::
        If a Runtime is not used with `ConnectionMode.DEFAULT` or
        `ConnectionMode.ALWAYS_ON_EXECUTE`, then external interference can cause undesired
        behaviour. Furthermore, the specifics of each mode might have slight variations
        depending on which Runtime is used (which should be well documented).
    """

    # Individual flags
    CONNECT_BEFORE_EXECUTE = auto()
    DISCONNECT_AFTER_EXECUTE = auto()
    CONNECT_AT_BEGINNING = auto()
    DISCONNECT_AT_END = auto()

    # Some pre-defined composite options
    MANUAL = auto()
    DEFAULT = CONNECT_BEFORE_EXECUTE | DISCONNECT_AFTER_EXECUTE
    ALWAYS = CONNECT_AT_BEGINNING | DISCONNECT_AT_END
    ALWAYS_ON_EXECUTE = CONNECT_BEFORE_EXECUTE | DISCONNECT_AT_END



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
import numpy as np

from qat.ir.gates.gates_1q import Hadamard, Phase, Rx, Ry, Rz, S, Sdg, T, Tdg, U, X, Y, Z
from qat.ir.gates.gates_2q import CNOT, ECR, SWAP, CPhase, CRx, CRy, CRz
from qat.ir.gates.operation import Barrier, Measure, Reset
from qat.ir.instructions import Instruction
from qat.ir.qat_ir import QatIR


class CircuitBuilder:
    """The :class:`CircuitBuilder` is used to construct high-level quantum programs.

    It can be used to implement high-level operations such as quantum gates and measure
    operations.
    """

    # TODO: support pulse instructions (COMPILER-304)
    # TODO: support named classical registers (COMPILER-285)
    # TODO: support control-flow; requires control flow instructions to be established
    # across all IR (COMPILER-146)

    def __init__(self, num_qubits: int, num_clbits: int = None):
        """
        :param num_qubits: The number of qubits in the circuit.
        :param num_clbits: The number of classical registers to save measurements too.
        """

        self.num_qubits = num_qubits
        self.num_clbits = num_clbits if num_clbits is not None else num_qubits
        self.instructions: list[Instruction] = []

    def add(self, gate):
        """Adds an operation to the circuit, validating the qubits indices are valid."""
        if not all([qubit < self.num_qubits for qubit in gate.qubits]):
            raise ValueError(
                f"The operation {gate} contains qubits with indices outside the valid "
                "range."
            )
        if len(gate.qubits) != len(set(gate.qubits)):
            raise ValueError(
                f"The operation {gate} contains qubits with duplicate indices."
            )
        self.instructions.append(gate)
        return self

    def X(self, qubit: int):
        r"""Adds an :math:`X` gate."""
        return self.add(X(qubit=qubit))

    def Y(self, qubit: int):
        r"""Adds a :math:`Y` gate."""
        return self.add(Y(qubit=qubit))

    def Z(self, qubit: int):
        r"""Adds a :math:`Z` gate."""
        return self.add(Z(qubit=qubit))

    def Rx(self, qubit: int, theta: float = np.pi):
        """Rotates a qubit around the x-axis by angle `theta`."""
        return self.add(Rx(qubit=qubit, theta=theta))

    def Ry(self, qubit: int, theta: float = np.pi):
        """Rotates a qubit around the y-axis by angle `theta`."""
        return self.add(Ry(qubit=qubit, theta=theta))

    def Rz(self, qubit: int, theta: float = np.pi):
        """Rotates a qubit around the z-axis by angle `theta`."""
        return self.add(Rz(qubit=qubit, theta=theta))

    def U(self, qubit: int, theta: float = 0.0, phi: float = 0.0, lambd: float = 0.0):
        """Apply a full rotation to a qubit around the Bloch sphere given three rotation
        angles.
        """
        return self.add(U(qubit=qubit, theta=theta, phi=phi, lambd=lambd))

    def Phase(self, qubit: int, theta: float = 0.0):
        r"""Applies a phase gate to a qubit, which can be considered equivalent to an
        :math:`R_{z}(\theta)` gate up to a global phase."""
        return self.add(Phase(qubit=qubit, theta=theta))

    def Hadamard(self, qubit: int):
        return self.add(Hadamard(qubit=qubit))

    def S(self, qubit: int):
        return self.add(S(qubit=qubit))

    def Sdg(self, qubit: int):
        return self.add(Sdg(qubit=qubit))

    def T(self, qubit: int):
        return self.add(T(qubit=qubit))

    def Tdg(self, qubit: int):
        return self.add(Tdg(qubit=qubit))

    def ECR(self, qubit1: int, qubit2: int):
        return self.add(ECR(qubit1=qubit1, qubit2=qubit2))

    def CNOT(self, control: int, target: int):
        return self.add(CNOT(qubit1=control, qubit2=target))

    def CX(self, control: int, target: int):
        return self.CNOT(control, target)

    def SWAP(self, qubit1: int, qubit2: int):
        return self.add(SWAP(qubit1=qubit1, qubit2=qubit2))

    def CRx(self, control: int, target: int, theta: float = np.pi):
        """Rotates the target qubit around the x-axis by angle `theta`, controlled by the
        state of another qubit."""
        return self.add(CRx(qubit1=control, qubit2=target, theta=theta))

    def CRy(self, control: int, target: int, theta: float = np.pi):
        """Rotates the target qubit around the y-axis by angle `theta`, controlled by the
        state of another qubit."""
        return self.add(CRy(qubit1=control, qubit2=target, theta=theta))

    def CRz(self, control: int, target: int, theta: float = np.pi):
        """Rotates the target qubit around the z-axis by angle `theta`, controlled by the
        state of another qubit."""
        return self.add(CRz(qubit1=control, qubit2=target, theta=theta))

    def CPhase(self, control: int, target: int, theta: float = np.pi):
        """Applies a phase rotation to a target qubit by angle `theta`, controlled by the
        state of another qubit."""
        return self.add(CPhase(qubit1=control, qubit2=target, theta=theta))

    def measure(self, qubit: int, clbit: int = None):
        """Measures a qubit and saves the result in the specified classical bit. If the
        classical bit isn't given, it will be decided automatically."""
        # TODO: implement a more sophisticed allocation procedure (COMPILER-305)
        if clbit is None:
            pass
        return self.add(Measure(qubit=qubit, clbit=clbit))

    def reset(self, qubit: int):
        """Reset a qubit to its :math:`|0>` state."""
        return self.add(Reset(qubit=qubit))

    def barrier(self, *qubits: int):
        """Adds a barrier between circuit operations to prevent optimisation over the
        barrier and sync qubits."""
        return self.add(Barrier(qubits=qubits))

    def emit(self):
        """Exports the circuit as :class:`QatIR`.

        Currently just wraps the instructions in :class:`QatIR`. Later the emitter could
        take on more responsibilities (e.g. some form of validation checks). It could prove
        benefitical to move this responsibility to a front end that wraps the builder as
        Qat IR complexity increases.
        """
        return QatIR(instructions=self.instructions)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024-2025 Oxford Quantum Circuits Ltd

from dataclasses import dataclass, field

from qat.core.metrics_base import MetricsManager, MetricsType
from qat.core.pass_base import AnalysisPass, ResultManager
from qat.core.result_base import ResultInfoMixin
from qat.purr.compiler.builders import InstructionBuilder
from qat.purr.compiler.devices import PulseChannel, Qubit, Resonator
from qat.purr.compiler.hardware_models import QuantumHardwareModel
from qat.purr.compiler.instructions import Acquire, CustomPulse, Pulse
from qat.purr.utils.logger import get_default_logger

log = get_default_logger()


@dataclass
class ActiveChannelResults(ResultInfoMixin):
    """Stores the active pulse channels in a task, which is defined by pulse channels that
    are acted on by a pulse or acquisition.

    Results are stored as a map between pulse channels and the qubit they belong to. Various
    helper properties and methods can be used to fetch the complete lists of active pulse
    channels and qubits.
    """

    target_map: dict[PulseChannel, Qubit] = field(default_factory=lambda: dict())

    @property
    def physical_qubit_indices(self) -> set[int]:
        """Returns a list of all active physical qubit indices."""
        return set([qubit.index for qubit in self.target_map.values()])

    @property
    def targets(self) -> list[PulseChannel]:
        """Returns a dictionary of all pulse channels with their full id as a key."""
        return list(self.target_map.keys())

    @property
    def qubits(self) -> list[Qubit]:
        """Returns a list of all active qubits."""
        return list(set(self.target_map.values()))

    def from_qubit(self, qubit: Qubit) -> list[PulseChannel]:
        """Returns the list of pulse channels that belong to a qubit."""
        pulse_channels = []
        for key, val in self.target_map.items():
            if val == qubit:
                pulse_channels.append(key)
        return pulse_channels


class ActivePulseChannelAnalysis(AnalysisPass):
    """Determines the set of pulse channels which are targeted by quantum instructions.

    A pulse channel that has a pulse played at any time, or an acquisition is defined to be
    active. This pass is used to determine which pulse channels are required in compilation,
    and is used in subsequent passes to easily extract pulse channel properties, and is
    useful for not performing extra analysis on rogue channels picked up by
    :class:`Synchronize` instructions.
    """

    # TODO: PydActivePulseChannelAnalysis: this will be even more useful for pydantic
    # instructions (COMPILER-393)

    def __init__(self, model: QuantumHardwareModel):
        self.model = model

    def run(
        self,
        ir: InstructionBuilder,
        res_mgr: ResultManager,
        met_mgr: MetricsManager,
        *args,
        **kwargs,
    ) -> InstructionBuilder:
        """
        :param ir: The list of instructions stored in an :class:`InstructionBuilder`.
        :param res_mgr: The result manager to store the analysis results.
        """

        targets: set[PulseChannel] = set()
        for inst in ir.instructions:
            if isinstance(inst, (Acquire, Pulse, CustomPulse)):
                targets.add(next(iter(inst.quantum_targets)))

        result = ActiveChannelResults()
        for target in targets:
            devices = self.model.get_devices_from_pulse_channel(target)

            if len(devices) == 0:
                devices = self.model.get_devices_from_physical_channel(
                    target.physical_channel
                )

            if len(devices) > 1:
                log.warning(
                    f"Multiple targets found with pulse channel {target}: "
                    + ", ".join([str(device) for device in devices])
                    + f". Defaulting to the first quantum device found, {devices[0]}."
                )

            device = devices[0]
            if isinstance(device, Resonator):
                device = self.model._map_resonator_to_qubit(device)

            if device:
                result.target_map[target] = device
            else:
                result.unassigned.append(target)

        phys_q_indices = sorted(list(result.physical_qubit_indices))
        log.info(f"Physical qubits used in this circuit: {phys_q_indices}")
        met_mgr.record_metric(MetricsType.PhysicalQubitIndices, phys_q_indices)

        res_mgr.add(result)
        return ir



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

import itertools
from typing import List, Tuple

import numpy as np
from more_itertools import partition

from qat.purr.backends.qblox.analysis_passes import BindingResult, TriageResult
from qat.purr.compiler.builders import InstructionBuilder
from qat.purr.compiler.hardware_models import QuantumHardwareModel
from qat.purr.compiler.instructions import (
    Acquire,
    EndRepeat,
    EndSweep,
    Instruction,
    Repeat,
    Return,
    Sweep,
)
from qat.purr.core.metrics_base import MetricsManager
from qat.purr.core.pass_base import TransformPass
from qat.purr.core.result_base import PreservedResults, ResultManager


class ScopeSanitisation(TransformPass):
    def run(self, ir: InstructionBuilder, res_mgr: ResultManager, *args, **kwargs):
        """
        Bubbles up all sweeps and repeats to the beginning of the list.
        Adds delimiter instructions to the repeats and sweeps signifying the end of their scopes.

        Intended for legacy existing builders and the relative order of instructions guarantees backwards
        compatibility.
        """

        tail, head = partition(
            lambda inst: isinstance(inst, (Sweep, Repeat)), ir.instructions
        )
        tail, head = list(tail), list(head)

        delimiters = [
            EndSweep() if isinstance(inst, Sweep) else EndRepeat() for inst in head
        ]

        ir.instructions = head + tail + delimiters[::-1]


class RepeatSanitisation(TransformPass):
    def __init__(self, model: QuantumHardwareModel = None):
        self.model = model

    def run(self, ir: InstructionBuilder, res_mgr: ResultManager, *args, **kwargs):
        """
        Fixes repeat instructions if any with default values from the HW model
        Adds a repeat instructions if none is found
        """

        repeats = [inst for inst in ir.instructions if isinstance(inst, Repeat)]
        if repeats:
            for rep in repeats:
                if rep.repeat_count is None:
                    rep.repeat_count = self.model.default_repeat_count
                if rep.repetition_period is None:
                    rep.repetition_period = self.model.default_repetition_period
        else:
            ir.repeat(self.model.default_repeat_count, self.model.default_repetition_period)


class ReturnSanitisation(TransformPass):
    def run(self, ir: InstructionBuilder, res_mgr: ResultManager, *args, **kwargs):
        returns = [inst for inst in ir.instructions if isinstance(inst, Return)]
        acquires = [inst for inst in ir.instructions if isinstance(inst, Acquire)]

        if returns:
            unique_variables = set(itertools.chain(*[ret.variables for ret in returns]))
            for ret in returns:
                ir.instructions.remove(ret)
        else:
            # If we don't have an explicit return, imply all results.
            unique_variables = set(acq.output_variable for acq in acquires)

        ir.returns(list(unique_variables))


class DesugaringPass(TransformPass):
    """Transforms syntactic sugars and implicit effects.

    The goal of this pass is to desugar any syntactic and semantic sugars. One of these
    sugars is iteration constructs such as :class:`Sweep` and :class:`Repeat`.

    1- Adds an iterator variable for sweep loops. We state that sweep loops implicitly
       declare an iterator variable to control progress of the sweep. See :class:`BindingPass`
    """

    def run(self, ir: InstructionBuilder, res_mgr: ResultManager, *args, **kwargs):
        """:param ir: The list of instructions stored in an :class:`InstructionBuilder`."""

        for inst in ir.instructions:
            if isinstance(inst, Sweep):
                count = len(next(iter(inst.variables.values())))
                iter_name = f"{hash(inst)}"
                inst.variables[iter_name] = np.linspace(1, count, count, dtype=int).tolist()


class ScopePeeling(TransformPass):
    """
    A transform pass to discard scopes assuming they’ve been identified.

    This pass is particularly useful for the code generator as it helps peel away any unwanted scopes.
    In fact, some sweeps are not lowerable and the user may choose to annotate certain sweeps and opt
    for static injection of variables with a switerator-like emitter. These sweep loops will be handled
    by the emitter statically while allowing it to lower nested constructs efficiently.
    """

    def run(
        self,
        ir: InstructionBuilder,
        res_mgr: ResultManager,
        met_mgr: MetricsManager,
        *args,
        **kwargs,
    ):
        scopes: List[Tuple[Instruction, Instruction]] = kwargs.get("scopes", [])

        if scopes is None or len(scopes) == 0:
            return PreservedResults.all()

        if any(not isinstance(scope, Tuple) or len(scope) != 2 for scope in scopes):
            raise ValueError(
                f"Invalid scopes argument. Expected a list of tuples of length 2, got {scopes}"
            )

        if any(s is None or e is None for (s, e) in scopes):
            raise ValueError(
                "Invalid scopes argument. Delimiter instructions must not be none."
            )

        # Results assumed to be correct at this level of analysis
        triage_result = res_mgr.lookup_by_type(TriageResult)
        binding_result = res_mgr.lookup_by_type(BindingResult)

        for target, instructions in triage_result.target_map.items():
            scoping_result = binding_result.scoping_results[target]
            for scope in scopes:
                if scope not in scoping_result.scope2symbols:
                    raise ValueError(
                        f"Could not find scope {scope}, the binding analysis result may have been corrupt"
                    )

        delimiters = set(itertools.chain(*scopes))

        def predicate(inst):
            return inst not in delimiters

        instructions = ir.splice()
        ir.instructions = list(filter(predicate, instructions))

        return PreservedResults.discard(triage_result, binding_result)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
import os
import tempfile

from compiler_config.config import (
    CompilerConfig,
    Languages,
    MetricsType,
    Qiskit,
    QiskitOptimizations,
    Tket,
    TketOptimizations,
    get_optimizer_config,
)
from qiskit import QuantumCircuit, qasm2, transpile
from qiskit.transpiler import TranspilerError

from qat.core.metrics_base import MetricsManager
from qat.core.pass_base import TransformPass
from qat.core.result_base import ResultManager
from qat.frontend.passes.analysis import InputAnalysisResult
from qat.ir.builder_factory import BuilderFactory
from qat.purr.compiler.builders import InstructionBuilder
from qat.purr.compiler.hardware_models import QuantumHardwareModel
from qat.purr.integrations.qasm import CloudQasmParser, Qasm3Parser
from qat.purr.integrations.tket import run_tket_optimizations_qasm


class FlattenIR(TransformPass):
    """Flatten the IR by removing nested structures like InstructionBlocks."""

    def run(self, ir: InstructionBuilder, *args, **kwargs):
        ir._instructions = ir.instructions
        return ir


class InputOptimisation(TransformPass):
    """Run third party optimisation passes on the incoming QASM."""

    def __init__(self, hardware: QuantumHardwareModel, *args, **kwargs):
        """Instantiate the pass with a hardware model.

        :param model: The hardware model is used in TKET optimisations.
        """
        self.hardware = hardware

    def run(
        self,
        program: str,
        res_mgr: ResultManager,
        met_mgr: MetricsManager,
        *args,
        compiler_config: CompilerConfig,
        **kwargs,
    ):
        """
        :param program: The program as a string (e.g. QASM or QIR), or filepath to the
            program.
        :param res_mgr: The results manager to look-up the :class:`InputAnalysisResults`.
        :param met_mgr: The metrics manager to save the optimised circuit.
        :param compiler_config: The compiler config should be provided by a keyword
            argument.
        """

        input_results = res_mgr.lookup_by_type(InputAnalysisResult)
        language = input_results.language
        program = input_results.raw_input
        if compiler_config.optimizations is None:
            compiler_config.optimizations = get_optimizer_config(language)
        if language in (Languages.Qasm2, Languages.Qasm3):
            program = self.run_qasm_optimisation(
                program, compiler_config.optimizations, met_mgr
            )
        return program

    def run_qasm_optimisation(self, qasm_string, optimizations, met_mgr, *args, **kwargs):
        """Extracted from DefaultOptimizers.optimize_qasm"""

        if (
            isinstance(optimizations, Tket)
            and optimizations.tket_optimizations != TketOptimizations.Empty
        ):
            qasm_string = run_tket_optimizations_qasm(
                qasm_string, optimizations.tket_optimizations, self.hardware
            )

        # TODO: [QK] Spend time looking at qiskit optimization and seeing if it's
        #   worth keeping around.
        if (
            isinstance(optimizations, Qiskit)
            and optimizations.qiskit_optimizations != QiskitOptimizations.Empty
        ):
            qasm_string = self.run_qiskit_optimization(
                qasm_string, optimizations.qiskit_optimizations
            )

        met_mgr.record_metric(MetricsType.OptimizedCircuit, qasm_string)
        return qasm_string

    def run_qiskit_optimization(self, qasm_string, level):
        """
        TODO: [QK] Current setup is unlikely to provide much benefit, refine settings
            before using.
        """
        if level is not None:
            try:
                optimized_circuits = transpile(
                    QuantumCircuit.from_qasm_str(qasm_string),
                    basis_gates=["u1", "u2", "u3", "cx"],
                    optimization_level=level,
                )
                qasm_string = qasm2.dumps(optimized_circuits)
            except TranspilerError:
                pass
                # log.warning(f"Qiskit transpile pass failed. {str(ex)}")

        return qasm_string


class Parse(TransformPass):
    """Parses the QASM/QIR input into IR."""

    def __init__(self, hardware: QuantumHardwareModel):
        """Instantiate the pass with a hardware model.

        :param model: The hardware model is required to create Qat IR.
        """
        self.hardware = hardware

    def run(
        self,
        program: str,
        res_mgr: ResultManager,
        *args,
        compiler_config: CompilerConfig,
        **kwargs,
    ):
        """
        :param program: The program as a string (e.g. QASM or QIR).
        :param res_mgr: The results manager to look-up the :class:`InputAnalysisResults`.
        :param compiler_config: The compiler config should be provided by a keyword
            argument.
        """
        input_results = res_mgr.lookup_by_type(InputAnalysisResult)
        language = input_results.language
        builder = self.hardware.create_builder()
        parser = None
        if language == Languages.QIR:
            builder = self.parse_qir(program, compiler_config)
        elif language == Languages.Qasm2:
            parser = CloudQasmParser()
        elif language == Languages.Qasm3:
            parser = Qasm3Parser()
        if parser is not None:
            if compiler_config.results_format.format is not None:
                parser.results_format = compiler_config.results_format.format
            builder = parser.parse(builder, program)

        return (
            self.hardware.create_builder()
            .repeat(
                compiler_config.repeats,
                repetition_period=compiler_config.repetition_period,
                passive_reset_time=compiler_config.passive_reset_time,
            )
            .add(builder)
        )

    def parse_qir(self, qir_string, compiler_config):
        """Extracted from QIRFrontend"""
        # TODO: Resolve circular import
        from qat.purr.integrations.qir import QIRParser

        # TODO: Remove need for saving to file before parsing
        suffix = ".bc" if isinstance(qir_string, bytes) else ".ll"
        with tempfile.NamedTemporaryFile(suffix=suffix, delete=False) as fp:
            if suffix == ".ll":
                fp.write(qir_string.encode())
            else:
                fp.write(qir_string)
            fp.close()
            try:
                builder = BuilderFactory.create_builder(self.hardware)
                parser = QIRParser(self.hardware, builder=builder)
                if compiler_config.results_format.format is not None:
                    parser.results_format = compiler_config.results_format.format
                quantum_builder = parser.parse(fp.name)
            finally:
                os.remove(fp.name)
        return quantum_builder



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

from qat.backend.passes.purr.analysis import (
    BindingPass,
    CFGPass,
    TILegalisationPass,
    TriagePass,
)
from qat.backend.passes.purr.transform import DesugaringPass
from qat.backend.qblox.codegen import QbloxBackend1, QbloxBackend2
from qat.backend.qblox.config.constants import QbloxTargetData
from qat.backend.qblox.passes.analysis import PreCodegenPass, QbloxLegalisationPass
from qat.core.pass_base import PassManager
from qat.frontend import AutoFrontend
from qat.middleend import CustomMiddleend
from qat.middleend.passes.purr.transform import (
    DeviceUpdateSanitisation,
    PhaseOptimisation,
    PostProcessingSanitisation,
    RepeatSanitisation,
    ReturnSanitisation,
    ScopeSanitisation,
)
from qat.middleend.passes.purr.validation import InstructionValidation, ReadoutValidation
from qat.model.target_data import TargetData
from qat.pipelines.pipeline import CompilePipeline
from qat.pipelines.updateable import PipelineConfig, UpdateablePipeline
from qat.purr.compiler.hardware_models import QuantumHardwareModel
from qat.purr.utils.logger import get_default_logger

log = get_default_logger()


def middleend_pipeline1(
    model: QuantumHardwareModel, target_data: TargetData
) -> PassManager:
    return (
        PassManager()
        | PhaseOptimisation()
        | PostProcessingSanitisation()
        | DeviceUpdateSanitisation()
        | InstructionValidation(target_data)
        | ReadoutValidation(model)
        | RepeatSanitisation(model, target_data)
        | ScopeSanitisation()
        | ReturnSanitisation()
        | DesugaringPass()
        | TriagePass()
        | BindingPass()
    )


def middleend_pipeline2(model: QuantumHardwareModel, target_data: TargetData):
    return (
        PassManager()
        | PhaseOptimisation()
        | PostProcessingSanitisation()
        | DeviceUpdateSanitisation()
        | InstructionValidation(target_data)
        | ReadoutValidation(model)
        | RepeatSanitisation(model, target_data)
        | ScopeSanitisation()
        | ReturnSanitisation()
        | DesugaringPass()
        | TriagePass()
        | BindingPass()
        | TILegalisationPass()
    )


def backend_pipeline1():
    return PassManager() | PreCodegenPass() | CFGPass()


def backend_pipeline2():
    return PassManager() | QbloxLegalisationPass() | PreCodegenPass() | CFGPass()


class QbloxCompilePipeline1(UpdateablePipeline):
    """A pipeline that compiles programs using the :class:`QbloxBackend1`.

    .. warning::

        This pipeline is for compilation purposes only and does not execute programs. Please
        select an appropriate execution pipeline if you wish to run the compiled programs.
    """

    @staticmethod
    def _build_pipeline(
        config: PipelineConfig,
        model: QuantumHardwareModel,
        target_data: QbloxTargetData | None,
        engine: None = None,
    ) -> CompilePipeline:
        if engine is not None:
            log.warning(
                "An engine was provided to the QbloxCompilePipeline, but it will "
                "be ignored. "
            )

        target_data = target_data if target_data is not None else QbloxTargetData.default()
        return CompilePipeline(
            model=model,
            target_data=target_data,
            frontend=AutoFrontend(model),
            middleend=CustomMiddleend(
                model=model,
                pipeline=middleend_pipeline1(model=model, target_data=target_data),
            ),
            backend=QbloxBackend1(
                model=model,
                pipeline=backend_pipeline1(),
            ),
            name=config.name,
        )


class QbloxCompilePipeline2(UpdateablePipeline):
    """A pipeline that compiles programs using the :class:`QbloxBackend2`.

    .. warning::

        This pipeline is for compilation purposes only and does not execute programs. Please
        select an appropriate execution pipeline if you wish to run the compiled programs.
    """

    @staticmethod
    def _build_pipeline(
        config: PipelineConfig,
        model: QuantumHardwareModel,
        target_data: QbloxTargetData | None,
        engine: None = None,
    ) -> CompilePipeline:
        if engine is not None:
            log.warning(
                "An engine was provided to the QbloxCompilePipeline, but it will "
                "be ignored. "
            )

        target_data = target_data if target_data is not None else QbloxTargetData.default()
        return CompilePipeline(
            model=model,
            target_data=target_data,
            frontend=AutoFrontend(model),
            middleend=CustomMiddleend(
                model=model,
                pipeline=middleend_pipeline2(model=model, target_data=target_data),
            ),
            backend=QbloxBackend2(
                model=model,
                pipeline=backend_pipeline2(),
            ),
            name=config.name,
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
from qat.backend.waveform.codegen import PydWaveformBackend
from qat.engines.waveform import EchoEngine
from qat.frontend import AutoFrontendWithFlattenedIR
from qat.middleend import PydDefaultMiddleend
from qat.model.hardware_model import PhysicalHardwareModel
from qat.model.target_data import TargetData
from qat.pipelines.pipeline import Pipeline
from qat.pipelines.updateable import PipelineConfig, UpdateablePipeline
from qat.purr.utils.logger import get_default_logger
from qat.runtime import SimpleRuntime
from qat.runtime.results_pipeline import get_results_pipeline

log = get_default_logger()


class EchoPipeline(UpdateablePipeline):
    """A pipeline that compiles programs using the :class:`PydWaveformBackend`
    and executes them using the :class:`EchoEngine`.

    An engine cannot be provided to the pipeline, as the EchoEngine is used directly.
    """

    @staticmethod
    def _build_pipeline(
        config: PipelineConfig,
        model: PhysicalHardwareModel,
        target_data: TargetData | None,
        engine: None = None,
    ) -> Pipeline:
        """Constructs a pipeline equipped with the :class:`PydWaveformBackend`
        and :class:`EchoEngine`."""

        if engine is not None:
            log.warning(
                "An engine was provided to the EchoPipeline, but it will be ignored. "
                "The EchoEngine is used directly."
            )

        target_data = target_data if target_data is not None else TargetData.default()
        results_pipeline = get_results_pipeline(model, target_data)
        return Pipeline(
            model=model,
            target_data=target_data,
            frontend=AutoFrontendWithFlattenedIR.default_for_pydantic(model),
            middleend=PydDefaultMiddleend(model, target_data),
            backend=PydWaveformBackend(model, target_data),
            runtime=SimpleRuntime(engine=EchoEngine(), results_pipeline=results_pipeline),
            name=config.name,
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

from collections import defaultdict
from warnings import warn

import numpy as np

from qat.backend.base import BaseBackend
from qat.backend.passes.purr.analysis import (
    IntermediateFrequencyAnalysis,
    IntermediateFrequencyResult,
    TimelineAnalysis,
    TimelineAnalysisResult,
)
from qat.backend.passes.purr.lowering import PartitionByPulseChannel
from qat.backend.passes.purr.validation import NoAcquireWeightsValidation
from qat.backend.waveform_v1.executable import (
    PositionalAcquireData,
    WaveformV1ChannelData,
    WaveformV1Program,
)
from qat.core.pass_base import InvokerMixin, MetricsManager, PassManager
from qat.core.result_base import ResultManager
from qat.executables import AcquireData, Executable
from qat.ir.instructions import Assign
from qat.ir.lowered import PartitionedIR
from qat.ir.measure import AcquireMode, PostProcessing
from qat.purr.backends.utilities import UPCONVERT_SIGN, evaluate_shape
from qat.purr.compiler.builders import InstructionBuilder
from qat.purr.compiler.devices import PulseChannel
from qat.purr.compiler.hardware_models import QuantumHardwareModel
from qat.purr.compiler.instructions import (
    CustomPulse,
    FrequencySet,
    FrequencyShift,
    Instruction,
    PhaseReset,
    PhaseSet,
    PhaseShift,
    Pulse,
    Reset,
    Waveform,
)
from qat.purr.utils.logger import get_default_logger

log = get_default_logger()


class WaveformV1Backend(BaseBackend[WaveformV1Program], InvokerMixin):
    """
    Target-machine code generation from an IR for targets that only require the explicit
    waveforms.
    """

    # TODO: replacing buffers calculations as passes: waveform generation pass, pulse
    # channel buffers pass, buffer amalgamation pass (COMPILER-413)

    def __init__(self, model: QuantumHardwareModel):
        """
        :param model: The hardware model that holds calibrated information on the qubits on the QPU.
                    As the emitter is used to generate code for some target machine, the hardware
                    model is needed for context-aware compilation.
        """
        warn(
            "WaveformV1 support is deprecated and will be removed in v4.",
            DeprecationWarning,
            stacklevel=2,
        )
        self.model = model

    def emit(
        self,
        ir: InstructionBuilder,
        res_mgr: ResultManager | None = None,
        met_mgr: MetricsManager | None = None,
        upconvert: bool = True,
        **kwargs,
    ) -> Executable[WaveformV1Program]:
        """Compiles :class:`InstructionBuilder` into an :class:`Executable`.

        Translates pulse instructions into explicit waveforms at the required times, and
        combines them across pulse channels to give a composite waveform on the necessary
        physical channels. Deals with low-level details such as pulse scheduling and phase
        shifts.
        """
        res_mgr = res_mgr if res_mgr is not None else ResultManager()
        met_mgr = met_mgr if met_mgr is not None else MetricsManager()
        ir = self.run_pass_pipeline(ir, res_mgr, met_mgr, **kwargs)
        timeline_res = res_mgr.lookup_by_type(TimelineAnalysisResult)
        if_res = res_mgr.lookup_by_type(IntermediateFrequencyResult)

        buffers = self.create_physical_channel_buffers(ir, timeline_res, upconvert)
        execute_acquire_dict, runtime_acquire_dict = self.create_acquires(ir, timeline_res)

        channels: dict[str, WaveformV1ChannelData] = {}
        for physical_channel, buffer in buffers.items():
            channels[physical_channel.full_id()] = WaveformV1ChannelData(
                baseband_frequency=if_res.frequencies.get(physical_channel, None),
                buffer=buffer,
                acquires=execute_acquire_dict.get(physical_channel, []),
            )

        returns = []
        for return_ in ir.returns:
            returns.extend(return_.variables)

        programs = self.create_programs(
            channels,
            ir.shots,
            ir.compiled_shots,
            timeline_res.total_duration,
        )

        return Executable[WaveformV1Program](
            programs=programs,
            acquires=runtime_acquire_dict,
            assigns=[Assign._from_legacy(assign) for assign in ir.assigns],
            returns=returns,
            calibration_id=self.model.calibration_id,
            shots=ir.shots,
        )

    def build_pass_pipeline(self, *args, **kwargs):
        return (
            PassManager()
            | NoAcquireWeightsValidation()
            | PartitionByPulseChannel()
            | TimelineAnalysis()
            | IntermediateFrequencyAnalysis(self.model)
        )

    @staticmethod
    def create_pulse_channel_buffer(
        pulse_channel: PulseChannel,
        instructions: list[Instruction],
        durations: list[int],
        upconvert: bool = True,
    ):
        """Creates a buffer of waveforms for a single pulse channel.

        :param pulse_channel: The pulse channel to create the buffer for.
        :param instructions: The list of instructions to process.
        :param durations: The list of durations for each instruction.
        :param upconvert: Whether to upconvert the waveforms to the target frequencies.
        """
        context = WaveformContext(pulse_channel, np.sum(durations))
        for i, instruction in enumerate(instructions):
            duration = durations[i]
            if isinstance(instruction, Waveform):
                context.process_pulse(instruction, duration, upconvert)
            elif isinstance(instruction, PhaseShift):
                context.process_phaseshift(instruction.phase)
            elif isinstance(instruction, PhaseSet):
                context.process_phaseset(instruction.phase)
            elif isinstance(instruction, PhaseReset):
                context.process_phasereset()
            elif isinstance(instruction, Reset):
                context.process_reset()
            elif isinstance(instruction, FrequencySet):
                context.process_frequencyset(instruction.frequency)
            elif isinstance(instruction, FrequencyShift):
                context.process_frequencyshift(instruction.frequency)
            else:
                context.process_delay(duration)
        return context.buffer

    def create_physical_channel_buffers(
        self,
        ir: PartitionedIR,
        timeline_res: TimelineAnalysisResult,
        upconvert: bool = True,
    ):
        """Creates a buffer of waveforms for each physical channel.

        :param ir: The partitioned IR containing the instructions.
        :param timeline_res: The timeline analysis result containing the number of samples
            for each instruction.
        :param upconvert: Whether to upconvert the waveforms to the target frequencies.
        """
        # Compute all pulse channel buffers
        pulse_channel_buffers = {}
        for pulse_channel in timeline_res.target_map:
            # The buffer is trivial if there are no waveforms
            if not any(
                [isinstance(inst, Waveform) for inst in ir.target_map[pulse_channel]]
            ):
                continue

            pulse_channel_buffers[pulse_channel] = self.create_pulse_channel_buffer(
                pulse_channel,
                ir.target_map[pulse_channel],
                timeline_res.target_map[pulse_channel].samples,
                upconvert,
            )

        # Organize by physical channels
        phys_to_pulse_map = {}
        for pulse_channel, buffer in pulse_channel_buffers.items():
            phys_to_pulse_map.setdefault(pulse_channel.physical_channel, []).append(buffer)

        # Compute sum of pulse channels
        physical_channel_buffers = {}
        for physical_channel in self.model.physical_channels.values():
            buffer_list = phys_to_pulse_map.get(physical_channel, [])
            if len(buffer_list) == 0:
                length = 0
            else:
                length = max(0, *[len(buffer) for buffer in buffer_list])
            buffer_sum = np.zeros(length, dtype=np.complex128)
            for buffer in buffer_list:
                buffer_sum[0 : len(buffer)] += buffer
            physical_channel_buffers[physical_channel] = buffer_sum

        return physical_channel_buffers

    def create_acquires(self, ir: PartitionedIR, timeline_res: TimelineAnalysisResult):
        """Assembles the acquire data needed to perform acquisition, and also the
        acquire information needed at runtime.

        :param ir: The partitioned IR containing the acquire map.
        :param timeline_res: The timeline analysis result containing the number of samples
            for each instruction.
        """
        execute_acquire_dict = defaultdict(list)
        runtime_acquire_dict = dict()
        for pulse_channel, acquire_list in ir.acquire_map.items():
            acq_list = execute_acquire_dict[pulse_channel.physical_channel]
            for acquire in acquire_list:
                idx = ir.target_map[pulse_channel].index(acquire)
                acq_list.append(
                    PositionalAcquireData(
                        length=timeline_res.target_map[pulse_channel].samples[idx],
                        position=np.sum(
                            timeline_res.target_map[pulse_channel].samples[0:idx]
                        ),
                        mode=acquire.mode,
                        output_variable=acquire.output_variable,
                    )
                )

                rp = ir.rp_map.get(acquire.output_variable, None)
                rp = rp.results_processing if rp is not None else None

                if acquire.mode == AcquireMode.INTEGRATOR:
                    shape = (ir.shots,)
                elif acquire.mode == AcquireMode.RAW:
                    shape = (
                        ir.shots,
                        timeline_res.target_map[pulse_channel].samples[idx],
                    )
                elif acquire.mode == AcquireMode.SCOPE:
                    shape = (timeline_res.target_map[pulse_channel].samples[idx],)

                runtime_acquire_dict[acquire.output_variable] = AcquireData(
                    mode=acquire.mode,
                    post_processing=[
                        PostProcessing._from_legacy(pp)
                        for pp in ir.pp_map.get(acquire.output_variable, [])
                    ],
                    shape=shape,
                    results_processing=rp,
                    physical_channel=pulse_channel.physical_channel.id,
                )

        return execute_acquire_dict, runtime_acquire_dict

    def create_programs(
        self,
        channel_data: dict[str, WaveformV1ChannelData],
        shots: int,
        batch_size: int,
        repetition_time: float,
    ) -> WaveformV1Program:
        """Creates WaveformV1Program instances from channel data.

        :param channel_data: The channel data for each physical channel.
        :param shots: The total number of shots to be executed.
        :param batch_size: The maximum number of shots per program.
        :param repetition_time: The time for each shot.
        """
        program = WaveformV1Program(
            channel_data=channel_data, repetition_time=repetition_time, shots=batch_size
        )
        programs = [program for _ in range(shots // batch_size)]
        if (remaining_shots := shots % batch_size) > 0:
            programs.append(
                WaveformV1Program(
                    channel_data=channel_data,
                    repetition_time=repetition_time,
                    shots=remaining_shots,
                )
            )
        return programs


class WaveformContext:
    """Used for waveform code generation for a particular pulse channel.

    This can be considered to be the dynamical state of a pulse channel which evolves as the
    circuit progresses.
    """

    # TODO: refactor this as a "channel backend" or as passes depending on when this is
    # done (COMPILER-413).

    def __init__(self, pulse_channel: PulseChannel, total_duration: int):
        """
        :param pulse_channel: The pulse channel that this is modelling.
        :param total_duration: The lifetime of the pulse channel in number of samples.
        """
        self.pulse_channel = pulse_channel
        self._buffer = np.zeros(total_duration, dtype=np.complex128)
        self._duration = 0
        self._phase = 0.0
        self._frequency = pulse_channel.frequency

    @property
    def physical_channel(self):
        return self.pulse_channel.physical_channel

    @property
    def buffer(self):
        return self._buffer

    def process_pulse(self, instruction: Pulse, samples: int, do_upconvert: bool = True):
        """Converts a waveform instruction into a discrete number of samples, handling
        upconversion to the target frequency if specified."""

        # TODO: the evaluate shape is handled in the EvaluatePulses pass, so this needs
        # adjusting to only accept square waveforms and custom pulses. (COMPILER-413)

        sample_time = self.pulse_channel.sample_time
        length = samples * sample_time
        centre = length / 2.0
        t = np.linspace(
            start=-centre + 0.5 * sample_time,
            stop=length - centre - 0.5 * sample_time,
            num=samples,
        )
        pulse = evaluate_shape(instruction, t, self._phase)

        scale = self.pulse_channel.scale
        if (
            isinstance(instruction, (Pulse, CustomPulse))
            and instruction.ignore_channel_scale
        ):
            scale = 1
        pulse *= scale
        pulse += self.pulse_channel.bias

        if do_upconvert:
            t += centre - 0.5 * sample_time + self._duration * sample_time
            pulse = self._do_upconvert(pulse, t)

        self._buffer[self._duration : self._duration + samples] = pulse
        self._duration += samples

    def _do_upconvert(self, buffer: list[float], time: list[float]):
        """A virtually NCO to upconvert the waveforms by a frequency that is the difference
        between the target frequency and the baseband frequency."""

        tslip = self.pulse_channel.phase_offset
        imbalance = self.pulse_channel.imbalance
        if self.pulse_channel.fixed_if:
            freq = self.pulse_channel.baseband_if_frequency
        else:
            freq = self._frequency - self.pulse_channel.baseband_frequency
        buffer *= np.exp(UPCONVERT_SIGN * 2.0j * np.pi * freq * time)
        if not tslip == 0.0:
            buffer_slip = buffer * np.exp(UPCONVERT_SIGN * 2.0j * np.pi * freq * tslip)
            buffer.imag = buffer_slip.imag
        if not imbalance == 1.0:
            buffer.real /= imbalance**0.5
            buffer.imag *= imbalance**0.5

        return buffer

    def process_phaseshift(self, phase: float):
        self._phase += phase

    def process_phaseset(self, phase: float):
        self._phase = phase

    def process_phasereset(self):
        self._phase = 0

    def process_reset(self):
        log.warning(
            "The WaveformV1Backend uses a `repetition_time` for resetting, so the reset "
            "instruction will be ignored."
        )

    def process_frequencyset(self, frequency: float):
        self._frequency = frequency

    def process_frequencyshift(self, frequency: float):
        self._frequency += frequency

    def process_delay(self, samples: int):
        self._duration += samples



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
from qat.pipelines.updateable import PipelineConfig

from .compile import LegacyCompilePipeline, middleend_pipeline
from .execute import LegacyExecutePipeline, LegacyPipelineConfig, results_pipeline
from .full import LegacyPipeline

__all__ = [
    "LegacyPipeline",
    "LegacyPipelineConfig",
    "LegacyCompilePipeline",
    "LegacyExecutePipeline",
    "PipelineConfig",
    "middleend_pipeline",
    "results_pipeline",
]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
from qat.model.loaders.base import BaseModelLoader


class CacheAccessLoader(BaseModelLoader):
    """A loader that loads a hardware model from a cache of hardware models.

    While primarily used for qatconfig, this loader can be applied to any cache that can be
    indexed by a hardware model name."""

    def __init__(self, cache, name: str):
        """
        :param cache: An object that can be indexed by a hardware model name, such as a
            dict.
        :param name: The name of the hardware model to load from the cache.
        """
        self._cache = cache
        self._name = name

    def load(self) -> BaseModelLoader:
        """Loads a hardware model from the cache."""
        return self._cache[self._name]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024-2025 Oxford Quantum Circuits Ltd
from __future__ import annotations

import numbers
import re
from collections.abc import Iterable
from copy import deepcopy
from functools import cached_property
from pydoc import locate
from typing import Annotated, Any, Type, TypeVar, Union, get_args, get_origin

import numpy as np
from frozendict import frozendict
from numpydantic import NDArray, Shape
from pydantic import (
    AfterValidator,
    BaseModel,
    BeforeValidator,
    ConfigDict,
    Field,
    PlainSerializer,
    PlainValidator,
    RootModel,
    computed_field,
    model_validator,
)
from pydantic._internal._model_construction import ModelMetaclass
from pydantic_core import core_schema

from qat.purr.utils.logger import get_default_logger

log = get_default_logger()


class NoExtraFieldsModel(BaseModel):
    """
    A Pydantic `BaseModel` with the extra constraints:
        #. Assignment of fields after initialisation is checked again.
        #. Extra fields given to the model are not ignored (default behaviour in `BaseModel`),
          but raise an error now.
    """

    model_config = ConfigDict(
        validate_assignment=True,
        use_enum_values=False,
        extra="forbid",
        ser_json_inf_nan="constants",
    )

    def __str__(self):
        return self.__repr__()


class NoExtraFieldsFrozenModel(NoExtraFieldsModel):
    """
    A Pydantic `BaseModel` with the extra constraints:
        #. Assignment of fields after initialisation is checked again.
        #. Extra fields given to the model are not ignored (default behaviour in `BaseModel`),
          but raise an error now.
        # All fields are frozen upon instantiation.
    """

    model_config = ConfigDict(frozen=True)


class AllowExtraFieldsModel(BaseModel):
    """
    A Pydantic `BaseModel` with the extra constraints:
        #. Assignment of fields after initialisation is checked again.
        #. Extra fields given to the model are ignored (default behaviour in `BaseModel`).
    """

    model_config = ConfigDict(
        validate_assignment=True, use_enum_values=False, extra="ignore"
    )

    def __str__(self):
        return self.__repr__()


class RehydratableModel(BaseModel):
    @computed_field
    @cached_property
    def object_type(self) -> str:
        """
        Returns the type of the object, which is the class name.
        """
        return self.__class__.__module__ + "." + self.__class__.__name__

    @classmethod
    def _rehydrate_object(cls, data):
        type_str = data.get("object_type", None)
        if type_str is None:
            cls_type = cls
        else:
            # Locate the class using the type string.
            cls_type = locate(type_str)
        if cls_type is None:
            raise ValueError(
                f"Could not locate class for type string '{type_str}'. Ensure it is a valid class path."
            )
        # Create an instance of the located class with the provided data.
        if not issubclass(cls_type, cls):
            raise TypeError(
                f"Located class '{cls_type.__name__}' is not a subclass of '{cls.__name__}'."
            )
        return cls_type(**data)


# This base file is used to implement classes/methods common to all hardware.


def validate_non_negative(value: int):
    if not isinstance(value, int) or value < 0:
        raise ValueError(f"Given value {value} must be an int and >=0.")
    return value


NonNegativeInt = Annotated[
    int,
    AfterValidator(validate_non_negative),
]


def validate_calibratable_positive_float(value: CalibratablePositiveFloat):
    if not np.isnan(value) and value < 0.0:
        raise ValueError(f"Given value {value} must be >=0.")
    return value


CalibratablePositiveFloat = Annotated[
    float,
    AfterValidator(validate_calibratable_positive_float),
]


def validate_calibratable_unit_interval(value: CalibratableUnitInterval):
    if not np.isnan(value):
        if value < 0.0 or value > 1.0:
            raise ValueError(f"Given value {value} must be in the interval [0, 1].")
    return value


# A float in the unit interval [0, 1].
CalibratableUnitInterval = Annotated[
    float,
    AfterValidator(validate_calibratable_unit_interval),
]


def validate_calibratable_unit_interval_array(array: CalibratableUnitInterval2x2Array):
    if np.any(array > 1) or np.any(array < 0):
        raise ValueError("Given array elements must be in the interval [0, 1].")
    return array


def validate_qubit_coupling(value: QubitCoupling):
    if isinstance(value, str):
        return tuple(map(int, re.findall(r"\d+", value)))
    elif isinstance(value, tuple):
        return value
    else:
        raise TypeError(
            "Invalid type for `QubitCoupling`. Please provide a `str` or `tuple`."
        )


# A qubit coupling represented as (q_i: int, q_j: int).
QubitCoupling = Annotated[
    tuple[NonNegativeInt, NonNegativeInt],
    BeforeValidator(validate_qubit_coupling),
]

VALIDATORS = BeforeValidator | AfterValidator | PlainValidator


K = TypeVar("GeneralKey")
V = TypeVar("GeneralValue")


class PydValidatedBase(RootModel):
    """
    Base class for validated Pydantic containers.
    """

    @model_validator(mode="after")
    def validation_setup(self):
        """
        Setup validation for the container.
        """
        annotation_args = get_args(self.__class__.model_fields["root"].annotation)
        value_type = annotation_args[0]
        annotation_type, allowed_types, validation_funcs = self._determine_validation_info(
            value_type
        )
        self._value_type = annotation_type
        self._value_types = allowed_types
        self._value_validators = validation_funcs

        return self

    @staticmethod
    def _determine_validation_info(annotation_type):
        validation_funcs = []
        if get_origin(annotation_type) is Annotated:
            annotation_type, *metadata = get_args(annotation_type)
            for item in metadata:
                if isinstance(item, VALIDATORS):
                    validation_funcs.append(item.func)
        if get_origin(annotation_type) is Union:
            allowed_types = get_args(annotation_type)
        else:
            allowed_types = (annotation_type,)
        return annotation_type, allowed_types, validation_funcs

    def validate_value(self, value: V):
        if not isinstance(value, self._value_types):
            raise TypeError(
                f"Cannot add value {value} of type '{type(value)}' to container of type {self._value_type}."
            )
        for validator in self._value_validators:
            value = validator(value)
        return value


class PydListBase(RootModel[list[V]]):
    root: list[V] = Field(default_factory=list)

    def __iter__(self):
        return iter(self.root)

    def __eq__(self, other: PydListBase | list):
        if isinstance(other, PydListBase):
            return self.root.__eq__(other.root)
        elif isinstance(other, list):
            return self.root.__eq__(other)
        return False

    def __getitem__(self, i: int):
        return self.root.__getitem__(i)

    def __len__(self):
        return self.root.__len__()

    def __repr__(self):
        return self.root.__repr__()

    def __str__(self):
        return self.root.__str__()

    def __deepcopy__(self, memo):
        copied_root = deepcopy(self.root, memo)
        copied_instance = self.__class__(root=copied_root)
        memo[id(self)] = copied_instance
        return copied_instance


class ValidatedList(PydListBase, PydValidatedBase):
    """
    A list object that validates the input appended/extended after instantiation.
    This way, we are sure that the elements in a list are only of a certain type.
    Pydantic containers only validate upon instantiation, not when modifying the
    container.
    """

    def append(self, value: V):
        self.validate_value(value)
        self.root.append(value)

    def extend(self, values: Iterable[V]):
        for value in values:
            self.append(value)

    def remove(self, value: V):
        self.root.remove(value)


def _validate_set(value: float | int | str | Iterable | None):
    if isinstance(value, (float, int, str)):
        value = {value}
    elif isinstance(value, (list, ValidatedList, tuple)):
        value = set(value)
    return value


class PydSetBase(RootModel[set[V]]):
    root: set[V] = Field(default_factory=set)

    def __iter__(self):
        return iter(self.root)

    def __eq__(self, other: PydSetBase):
        if isinstance(other, PydSetBase):
            return self.root.__eq__(other.root)
        elif isinstance(other, set):
            return self.root.__eq__(other)
        elif isinstance(other, (list, tuple)):
            return self.root.__eq__(set(other))
        return False

    def __le__(self, other: PydSetBase):
        if isinstance(other, PydSetBase):
            return self.root.issubset(other.root)
        elif isinstance(other, set):
            return self.root.issubset(other)
        else:
            raise NotImplementedError(f"Unsupported processing for {self} and {other}.")

    def __sub__(self, other: PydSetBase):
        if isinstance(other, PydSetBase):
            return self.root.__sub__(other.root)
        elif isinstance(other, set):
            return self.root.__sub__(other)
        else:
            raise NotImplementedError(f"Unsupported processing for {self} and {other}.")

    def __and__(self, other: PydSetBase):
        if isinstance(other, PydSetBase):
            return self.root.__and__(other.root)
        elif isinstance(other, set):
            return self.root.__and__(other)
        else:
            raise NotImplementedError(f"Unsupported processing for {self} and {other}.")

    def __len__(self):
        return self.root.__len__()

    def __repr__(self):
        return self.root.__repr__()

    def __str__(self):
        return self.root.__str__()

    def __deepcopy__(self, memo):
        copied_root = deepcopy(self.root, memo)
        copied_instance = self.__class__(root=copied_root)
        memo[id(self)] = copied_instance
        return copied_instance


class FrozenSet(PydSetBase):
    """
    A Pydantic set that is immutable after instantiation.
    """

    root: frozenset[V] = Field(default_factory=frozenset)


class ValidatedSet(PydSetBase, PydValidatedBase):
    """
    A set object that validates the input added after instantiation.
    This way, we are sure that the elements in a set are only of a certain type.
    Pydantic containers only validate upon instantiation, not when modifying the
    container.
    """

    def add(self, value: V):
        self.validate_value(value)
        self.root.add(value)

    def discard(self, value):
        self.root.discard(value)

    def pop(self):
        self.root.pop()

    def remove(self, value):
        self.root.remove(value)

    def update(self, *sets):
        for s in sets:
            for value in s:
                self.add(value)


class PydDictBase(RootModel[dict[K, V]]):
    root: dict[K, V] = Field(default_factory=dict)

    def get(self, key, default=None):
        return self.root.get(key, default)

    def keys(self):
        return self.root.keys()

    def items(self):
        return self.root.items()

    def values(self):
        return self.root.values()

    def __eq__(self, other: PydDictBase):
        if isinstance(other, PydDictBase):
            return self.root.__eq__(other.root)
        elif isinstance(other, dict):
            return self.root.__eq__(other)
        return False

    def __getitem__(self, key):
        return self.root.get(key, None)

    def pop(self, key: Any, *args, **kwargs):
        return self.root.pop(key, *args, **kwargs)

    def __iter__(self):
        return iter(self.root)

    def __len__(self):
        return self.root.__len__()

    def __repr__(self):
        return self.root.__repr__()

    def __str__(self):
        return self.root.__str__()

    def __deepcopy__(self, memo):
        result = {}
        for k, v in self.root.items():
            result[deepcopy(k)] = deepcopy(v)

        result = self.__class__(result)
        memo[id(self)] = result
        return result


class _PydanticFrozenDictAnnotation:
    """
    Helper class since Pydantic `V2` does only offer support for `frozenset`, not `frozendict`.
    """

    @classmethod
    def __get_pydantic_core_schema__(cls, source_type, handler):
        def validate_from_dict(d: dict | frozendict) -> frozendict:
            return frozendict(d)

        k, v = get_args(source_type)
        frozendict_schema = core_schema.chain_schema(
            [
                handler.generate_schema(dict[k, v]),
                core_schema.no_info_plain_validator_function(validate_from_dict),
                core_schema.is_instance_schema(frozendict),
            ]
        )
        return core_schema.json_or_python_schema(
            json_schema=frozendict_schema,
            python_schema=frozendict_schema,
            serialization=core_schema.plain_serializer_function_ser_schema(dict),
        )


pyd_frozendict = Annotated[frozendict[K, V], _PydanticFrozenDictAnnotation]


class FrozenDict(PydDictBase):
    """
    A Pydantic dict that is immutable after instantiation.
    """

    root: pyd_frozendict[K, V] = Field(default_factory=frozendict)


class ValidatedDict(PydDictBase, PydValidatedBase):
    """
    A dict object that validates the input added after instantiation.
    This way, we are sure that the elements in a dict are only of a certain type.
    Pydantic containers only validate upon instantiation, not when modifying the
    container.
    """

    @model_validator(mode="after")
    def validation_setup(self):
        """
        Validate the types of keys and values in the dictionary.
        """
        annotation_args = get_args(self.__class__.model_fields["root"].annotation)
        key_type = annotation_args[0]
        annotation_type, allowed_types, validation_funcs = self._determine_validation_info(
            key_type
        )
        self._key_type = annotation_type
        self._key_types = allowed_types
        self._key_validators = validation_funcs
        value_type = annotation_args[1]
        annotation_type, allowed_types, validation_funcs = self._determine_validation_info(
            value_type
        )
        self._value_type = annotation_type
        self._value_types = allowed_types
        self._value_validators = validation_funcs
        return self

    def validate_key(self, key: K):
        if not isinstance(key, self._key_types):
            raise TypeError(
                f"Cannot add key {key} of type '{type(key)}' to container of type {self._key_type}."
            )
        for validator in self._key_validators:
            key = validator(key)
        return key

    def update(self, data: dict[K, V]):
        for key, value in data.items():
            self.__setitem__(key, value)

    def __setitem__(self, key: K, value: V):
        self.validate_key(key)
        self.validate_value(value)
        self.root[key] = value


QubitId = NonNegativeInt


def validate_waveform_type(value: BaseModel):
    if isinstance(value, str):
        return locate("qat.ir.waveforms." + value)
    elif isinstance(value, ModelMetaclass):
        return value
    else:
        raise TypeError(
            "Invalid type for `WaveformType`. Please provide a `str` or `BaseModel`."
        )


WaveformType = Annotated[
    TypeVar("Waveform"),
    PlainSerializer(lambda wf: wf.__name__, when_used="json-unless-none"),
    BeforeValidator(validate_waveform_type),
]


def find_all_subclasses(cls: Type) -> list[Type]:
    """
    Recursively finds nested subclasses of a class.
    """
    subclasses = cls.__subclasses__()
    for subclass in subclasses:
        subclasses.extend(find_all_subclasses(subclass))
    return subclasses


# Efficient serializing of numeric numpy arrays
def _validate_value(
    value: str | list | np.ndarray,
    implied_type: np.dtype,
    required_type: np.dtype,
):
    """
    If value is a string: Reverts the hex value and type information into a numpy array.
    If value is a list or a numpy array: Make a numpy array and validate its type against ty
    """

    if isinstance(value, str):
        value = np.frombuffer(bytearray.fromhex(value), dtype=implied_type)
    elif isinstance(value, (list, np.ndarray)):
        value = np.asarray(value, dtype=implied_type)
    else:
        raise ValueError(
            f"Expected value to be {str | list | np.ndarray}, got {type(value)}"
        )

    if np.can_cast(value.dtype, required_type):
        return value.astype(required_type)

    try:
        if np.all((cast_value := value.astype(required_type)) == value):
            return cast_value
    except Exception as e:
        raise ValueError(f"""Cannot cast {value.dtype} to {required_type}\n{str(e)}""")


def _serializer(obj, ty: type):
    """Lists of complex numbers can be expensive to serialise: by serializing type
    information and its value as a hex, we can have more performant serialization."""

    dtype = np.dtype(ty)
    if isinstance(obj, PydArray):
        obj.value = obj.value.astype(dtype)
        return {
            "dtype": obj.value.dtype.name,
            "shape": obj.value.shape,
            "value": obj.value.tobytes().hex(),
        }
    elif isinstance(obj, np.ndarray):
        obj = obj.astype(dtype)
        return {"dtype": obj.dtype.name, "shape": obj.shape, "value": obj.tobytes().hex()}
    else:
        raise ValueError(f"Expected obj to be {PydArray | np.ndarray}, got {type(obj)}")


def _validator(payload, ty: type):
    """
    Plain validator function for annotated numpy array types.
    The payload is assumed to be consumed as a string, a numpy array, or a dictionary
    and a PydArray is created from it.
    """

    if isinstance(payload, PydArray):
        return payload
    elif isinstance(payload, np.ndarray):
        payload = _validate_value(
            payload, implied_type=payload.dtype, required_type=np.dtype(ty)
        )
        payload = {"dtype": payload.dtype, "shape": payload.shape, "value": payload}
    elif isinstance(payload, (str, list)):
        payload = _validate_value(
            payload, implied_type=np.dtype(ty), required_type=np.dtype(ty)
        )
        payload = {"dtype": payload.dtype, "shape": payload.shape, "value": payload}
    elif isinstance(payload, dict):
        dtype = payload.get("dtype", None)
        dtype = np.dtype(dtype) if dtype is not None else None
        shape = payload.get("shape", (0,))
        shape = tuple(shape)
        value = payload.get("value", [])
        value = _validate_value(value, implied_type=dtype, required_type=np.dtype(ty))
        payload = {"dtype": dtype, "shape": shape, "value": value}
    else:
        raise ValueError(
            f"Expected payload to be {str | list | np.ndarray | dict} got {type(payload)}"
        )

    value = payload["value"]
    shape = payload["shape"]
    return value.reshape(shape)


def annotate_pyd_array(ty: type):
    """Creates an annotated type for numeric lists with Pydantic serializers and validators
    for efficient serialization."""
    return Annotated[
        PydArray,  # TODO - Encode custom data type COMPILER-769
        PlainValidator(lambda x: _validator(x, ty)),
        PlainSerializer(lambda x: _serializer(x, ty)),
    ]


class PydArray(NoExtraFieldsModel, np.lib.mixins.NDArrayOperatorsMixin):
    """
    A data class wrapper to handle the information needed to completely describe a numpy array:
        + Value is the numpy array itself
        + Shape is the shape of the numpy array
        + dtype is the data (implied) type of the numpy array

    Through annotations, this allows creation of metadata classes on top of PydArray that describe
    how to (de)serialise a (blob) PydArray object according to some (required) type. See _validator()
    and _serializer().

    Enlisting the array data as (nested) list(s) is not optimal, and this class renders (de)serialisation
    fast.

    This class also mixes in NDArrayOperatorsMixin, which defines Python arithmetic operators in terms of
    NumPy ufuncs. This does NOT guarantee full interoperability with NumPy but allows the option to extend
    this support in the future. For now, the sole purpose of this class is ONLY to be a data class wrapper
    and mediator for (de)serialisation purposes.
    """

    value: NDArray[Shape["*, ..."], int | float | complex | bool]  # noqa: F722
    _HANDLED_TYPES = (np.ndarray, numbers.Number)

    def __init__(self, *args, **kwargs):
        if args:
            if len(args) != 1:
                raise TypeError(
                    f"{type(self).__name__} accepts at most 1 positional argument ('value'), got {len(args)}."
                )

            if "value" in kwargs:
                raise TypeError("Pass either a positional value or 'value=', not both")

            kwargs["value"] = args[0]

        kwargs["value"] = (
            np.asarray(kwargs["value"])
            if (isinstance(lst := kwargs.get("value", []), list))
            else lst
        )
        super().__init__(**kwargs)

    def __array__(self, dtype=None, copy=None) -> np.ndarray:
        dtype = dtype or self.dtype
        return self.value.astype(dtype=dtype, copy=copy)

    # One might also consider adding the built-in list type to this
    # list, to support operations like np.add(array_like, list)

    @staticmethod
    def _wrap(x):
        """Wrap ndarrays as PydArray; leave scalars/others as-is."""
        return PydArray(x) if isinstance(x, np.ndarray) else x

    def __array_function__(self, func, types, args, kwargs):
        # Only handle if any argument is of PydArray type
        if not any(issubclass(t, PydArray) for t in types):
            return NotImplemented

        # Unwrap all PydArray to ndarrays
        def unwrap(a):
            return a.value if isinstance(a, PydArray) else a

        uargs = tuple(unwrap(a) for a in args)
        ukwargs = {k: unwrap(v) for k, v in (kwargs or {}).items()}

        out = func(*uargs, **ukwargs)
        if isinstance(out, tuple):
            return tuple(self._wrap(x) for x in out)
        return self._wrap(out)

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        """
        Implements special methods for almost all of Python's built-in operators.

        Implementation inspired by
        https://numpy.org/doc/2.2/reference/generated/numpy.lib.mixins.NDArrayOperatorsMixin.html
        """
        out = kwargs.get("out", ())
        for x in inputs + out:
            # Use PydArray instead of type(self)
            # for isinstance to allow subclasses that don't
            # override __array_ufunc__ to handle `PydArray` objects.
            if not isinstance(x, (np.ndarray, numbers.Number, PydArray)):
                return NotImplemented
        # Defer to the implementation of the ufunc
        # on unwrapped values.
        inputs = tuple(x.value if isinstance(x, PydArray) else x for x in inputs)
        if out:
            kwargs["out"] = tuple(x.value if isinstance(x, PydArray) else x for x in out)

        result = getattr(ufunc, method)(*inputs, **kwargs)

        if type(result) is tuple:
            # multiple return values
            return tuple(type(self)(x) for x in result)
        elif method == "at":
            # ufunc.at performs in-place updates and returns None by design
            return None
        else:
            # one return value (default)
            return self._wrap(result)

    # Custom equality and inequality operators to improve compatibility of
    # Pydantic with empty arrays.
    def __eq__(self, other):
        if isinstance(other, PydArray):
            return np.array_equal(self.value, other.value)
        elif isinstance(other, np.ndarray):
            return np.array_equal(self.value, other)

        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    # Other niceties from the ndarray interface we'd like to support for PydArrays.
    def __getattr__(self, name):
        """
        Delegate attribute access to the underlying ndarray.
        This will handle .shape, .size, .reshape, etc. automatically.
        """
        attr = getattr(self.value, name)
        # If the attribute is callable (e.g. reshape), wrap its return value
        if callable(attr):

            def wrapper(*args, **kwargs):
                result = attr(*args, **kwargs)
                return self._wrap(result)

            return wrapper
        return attr

    def __getitem__(self, index):
        return self._wrap(self.value[index])

    def __setitem__(self, index, value):
        self.value[index] = value.value if isinstance(value, PydArray) else value

    def __len__(self):
        return len(self.value)

    def __iter__(self):
        for x in self.value:
            yield self._wrap(x)

    def __repr__(self):
        return "%s(%r, dtype=%r)" % (type(self).__name__, self.value, self.dtype)


IntNDArray = annotate_pyd_array(int)
FloatNDArray = annotate_pyd_array(float)
ComplexNDArray = annotate_pyd_array(complex)

CalibratableUnitInterval2x2Array = Annotated[
    PydArray,
    AfterValidator(validate_calibratable_unit_interval_array),
    PlainSerializer(lambda x: _serializer(x, float)),
]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024-2025 Oxford Quantum Circuits Ltd
from numbers import Number
from typing import List

import numpy as np

from qat.core.config.configure import get_config
from qat.purr.backends.live import LiveHardwareModel
from qat.purr.compiler.builders import InstructionBuilder
from qat.purr.compiler.devices import MaxPulseLength, PulseChannel, Qubit
from qat.purr.compiler.execution import QuantumExecutionEngine
from qat.purr.compiler.hardware_models import QuantumHardwareModel
from qat.purr.compiler.instructions import (
    Acquire,
    AcquireMode,
    CustomPulse,
    PostProcessing,
    ProcessAxis,
    Pulse,
    Sweep,
    Variable,
)
from qat.purr.core.pass_base import ValidationPass
from qat.purr.core.result_base import ResultManager


class InstructionValidation(ValidationPass):
    """
    Extracted from QuantumExecutionEngine.validate()
    """

    def __init__(
        self,
        engine: QuantumExecutionEngine,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.engine = engine

    def run(self, ir: InstructionBuilder, res_mgr: ResultManager, *args, **kwargs):
        qatconfig = get_config()

        instruction_length = len(ir.instructions)
        if instruction_length > self.engine.max_instruction_len:
            raise ValueError(
                f"Program too large to be run in a single block on current hardware. "
                f"{instruction_length} instructions."
            )

        for inst in ir.instructions:
            if isinstance(inst, Acquire) and not inst.channel.acquire_allowed:
                raise ValueError(
                    f"Cannot perform an acquire on the physical channel with id "
                    f"{inst.channel.physical_channel}"
                )
            if isinstance(inst, (Pulse, CustomPulse)):
                duration = inst.duration
                if isinstance(duration, Number) and duration > MaxPulseLength:
                    if (
                        qatconfig.INSTRUCTION_VALIDATION.PULSE_DURATION_LIMITS
                    ):  # Do not throw error if we specifically disabled the limit checks.
                        # TODO: Add a lower bound for the pulse duration limits as well in a later PR,
                        # which is specific to each hardware model and can be stored as a member variables there.
                        raise ValueError(
                            f"Max Waveform width is {MaxPulseLength} s "
                            f"given: {inst.duration} s"
                        )
                elif isinstance(duration, Variable):
                    values = next(
                        iter(
                            [
                                sw.variables[duration.name]
                                for sw in ir.instructions
                                if isinstance(sw, Sweep)
                                and duration.name in sw.variables.keys()
                            ]
                        )
                    )
                    if np.max(values) > MaxPulseLength:
                        if qatconfig.INSTRUCTION_VALIDATION.PULSE_DURATION_LIMITS:
                            raise ValueError(
                                f"Max Waveform width is {MaxPulseLength} s "
                                f"given: {values} s"
                            )


class ReadoutValidation(ValidationPass):
    """
    Extracted from LiveDeviceEngine.validate()
    """

    def __init__(
        self,
        model: QuantumHardwareModel,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.model = model

    def run(self, ir: InstructionBuilder, res_mgr: ResultManager, *args, **kwargs):
        qatconfig = get_config()

        if not isinstance(self.model, LiveHardwareModel):
            return

        consumed_qubits: List[str] = []
        chanbits_map = {}
        for inst in ir.instructions:
            if isinstance(inst, PostProcessing):
                if (
                    inst.acquire.mode == AcquireMode.SCOPE
                    and ProcessAxis.SEQUENCE in inst.axes
                ):
                    raise ValueError(
                        "Invalid post-processing! Post-processing over SEQUENCE is "
                        "not possible after the result is returned from hardware "
                        "in SCOPE mode!"
                    )
                elif (
                    inst.acquire.mode == AcquireMode.INTEGRATOR
                    and ProcessAxis.TIME in inst.axes
                ):
                    raise ValueError(
                        "Invalid post-processing! Post-processing over TIME is not "
                        "possible after the result is returned from hardware in "
                        "INTEGRATOR mode!"
                    )
                elif inst.acquire.mode == AcquireMode.RAW:
                    raise ValueError(
                        "Invalid acquire mode! The live hardware doesn't support "
                        "RAW acquire mode!"
                    )

            # Check if we've got a measure in the middle of the circuit somewhere.
            elif qatconfig.INSTRUCTION_VALIDATION.NO_MID_CIRCUIT_MEASUREMENT:
                if isinstance(inst, Acquire):
                    for qbit in self.model.qubits:
                        if qbit.get_acquire_channel() == inst.channel:
                            consumed_qubits.append(qbit)
                elif isinstance(inst, Pulse):
                    # Find target qubit from instruction and check whether it's been
                    # measured already.
                    acquired_qubits = [
                        (
                            (
                                chanbits_map[chanbit]
                                if chanbit in chanbits_map
                                else chanbits_map.setdefault(
                                    chanbit,
                                    self.model._resolve_qb_pulse_channel(chanbit)[0],
                                )
                            )
                            in consumed_qubits
                        )
                        for chanbit in inst.quantum_targets
                        if isinstance(chanbit, (Qubit, PulseChannel))
                    ]

                    if any(acquired_qubits):
                        raise ValueError(
                            "Mid-circuit measurements currently unable to be used."
                        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
from compiler_config.config import CompilerConfig

from qat.core.metrics_base import MetricsManager
from qat.core.result_base import ResultManager
from qat.frontend.base import BaseFrontend
from qat.frontend.fallthrough import FallthroughFrontend
from qat.frontend.passes.purr.transform import FlattenIR
from qat.frontend.passes.transform import PydFlattenIR
from qat.frontend.purr.purr import PurrFrontend
from qat.frontend.qasm import Qasm2Frontend, Qasm3Frontend
from qat.frontend.qat_ir import QatFrontend
from qat.frontend.qir import QIRFrontend
from qat.model.hardware_model import PhysicalHardwareModel as PydHardwareModel
from qat.purr.compiler.hardware_models import QuantumHardwareModel
from qat.utils.hardware_model import check_type_legacy_or_pydantic


class AutoFrontend(BaseFrontend):
    """Automatically deploys a frontend to match the type of source language.

    Provided with a number of frontends, the :class:`AutoFrontend` will attempt to match
    a source program to the frontend. If successful, it will use the matched frontend to
    emit the source program as Qat IR.

    .. warning::
        The :class:`AutoFrontend` will check the source programs against the frontends in
        the order provided. If multiple frontends are capable of processing a source file,
        then the order should be strategically chosen to have the desired outcome. For
        example, if using the :class:`FallthroughFrontend`, this should be placed at the
        end of line to allow other frontends the opportunity to be matched.
    """

    def __init__(
        self,
        model: QuantumHardwareModel | PydHardwareModel | None,
        *frontends: BaseFrontend,
    ):
        """
        :param model: The hardware model is needed to instantiate the default frontends,
            defaults to None.
        :param frontends: The different frontends to check the source file against.
        """

        self.model = check_type_legacy_or_pydantic(model)
        if len(frontends) == 0:
            frontends = self.default_frontends(model)
        self.frontends = frontends

    @staticmethod
    def default_frontends(model) -> list[BaseFrontend]:
        """Returns a default list of frontends to try.

        :param model: The hardware model is needed to instantiate the default frontends.
        """
        return [Qasm2Frontend(model), Qasm3Frontend(model), QIRFrontend(model)]

    @classmethod
    def default_for_purr(cls, model: QuantumHardwareModel) -> "AutoFrontend":
        """Temporary factory for creating an AutoFrontend that allows a PurrFrontend.

        :param model: The purr hardware model is needed to instantiate the default
            frontends.
        """

        # TODO: remove when purr is fully deprecated (COMPILER-724)
        frontends = cls.default_frontends(model) + [PurrFrontend(model)]
        return cls(model, *frontends)

    @classmethod
    def default_for_pydantic(cls, model: PydHardwareModel) -> "AutoFrontend":
        """Temporary factory for creating an AutoFrontend that allows a QatFrontend.

        :param model: The hardware model is needed to instantiate the default
            frontends.
        """

        # TODO: The default_frontends methods should eventually be replaced with this.
        # (COMPILER-724)
        frontends = cls.default_frontends(model) + [QatFrontend(model)]
        return cls(model, *frontends)

    @classmethod
    def default_for_legacy(cls, model: QuantumHardwareModel) -> "AutoFrontend":
        """Temporary factory for creating an AutoFrontend that has a fallthrough frontend
        to allow any legacy source."""

        # TODO: remove when legacy is fully deprecated (COMPILER-724)
        frontends = cls.default_frontends(model) + [FallthroughFrontend(model)]
        return cls(model, *frontends)

    def check_and_return_source(self, src):
        """Checks the source file sequentially against the different frontends. If a
        suitable frontend is found, the source program is returned.

        :param src: The source program, or path to the program.
        :returns: If the program is determined to not be valid, False is returned.
            Otherwise, the program is returned (and loaded if required).
        """
        assigned_frontend = self.assign_frontend(src)
        if assigned_frontend is None:
            return False
        return assigned_frontend.check_and_return_source(src)

    def assign_frontend(self, src) -> BaseFrontend | None:
        """Determines the source language and assigns the correct frontend.

        :param src: The source program to discriminate.
        :returns: The first frontend that is able to recognise the source program. Returns
            None if no suitable frontend is found.
        """

        for frontend in self.frontends:
            if frontend.check_and_return_source(src):
                return frontend
        return None

    def emit(
        self,
        src,
        res_mgr: ResultManager | None = None,
        met_mgr: MetricsManager | None = None,
        compiler_config: CompilerConfig | None = None,
        **kwargs,
    ):
        """Chooses a suitable frontend and compiles a given source program.

        :param src: The source program, or path to the source program.
        :param res_mgr: The result manager to store results from passes within the pipeline,
            defaults to None.
        :param met_mgr: The metrics manager to store metrics, such as the optimized QASM
            circuit, defaults to None.
        :param compiler_config: The compiler config, defaults to None.
        :raises ValueError: An error is thrown if no suitable frontend could be found.
        """

        for frontend in self.frontends:
            if frontend.check_and_return_source(src):
                return frontend.emit(src, res_mgr, met_mgr, compiler_config, **kwargs)
        raise ValueError("No suitable frontend could be found for the source program.")


class AutoFrontendWithFlattenedIR(AutoFrontend):
    """
    An :class:`AutoFrontend` that flattens the IR after parsing.
    """

    def emit(
        self,
        src,
        res_mgr: ResultManager | None = None,
        met_mgr: MetricsManager | None = None,
        compiler_config: CompilerConfig | None = None,
        **kwargs,
    ):
        ir = super().emit(src, res_mgr, met_mgr, compiler_config, **kwargs)

        if self.model and isinstance(self.model, PydHardwareModel):
            return PydFlattenIR().run(ir)
        return FlattenIR().run(ir)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

from qat.core.metrics_base import MetricsManager
from qat.core.pass_base import PassManager
from qat.core.result_base import ResultManager
from qat.engines import NativeEngine
from qat.executables import Executable
from qat.runtime import BaseRuntime
from qat.runtime.aggregator import QBloxAggregator, ResultsAggregator
from qat.runtime.connection import ConnectionMode


class SimpleRuntime(BaseRuntime):
    """
    The entry point to the execution interface. Programs generated from the backend are wrapped
    in a :class:Executable object which further specifies high level metadata about acquisition
    restrictions such as the expected shape of any loop nest, the post-processing, and results
    formatting. The runtime provides the following services.

    - Execution batching: Some programs require memory specifications beyond that of the control
      hardware. It important to slice and batch-execute large programs. Equally important is
      necessity to aggregate results from different batches.
    - Postprocessing: Performs any required post-processing steps that haven't been carried out
      real-time on the FPGA.
    - Error mitigation: Adjusts results based on error mitigation strategies.

    These services are specified as a pipline of passes, see `results_pipeline` module.
    """

    def __init__(
        self,
        engine: NativeEngine,
        results_pipeline: PassManager | None = None,
        connection_mode: ConnectionMode = ConnectionMode.DEFAULT,
        aggregator: ResultsAggregator | QBloxAggregator | None = None,
    ):
        super().__init__(engine, results_pipeline, connection_mode)
        self.aggregator = aggregator or ResultsAggregator()

    def execute(
        self,
        executable: Executable,
        res_mgr: ResultManager | None = None,
        met_mgr: MetricsManager | None = None,
        **kwargs,
    ):
        """Fully execute a package against the hardware with batching of shots and results
        post-processing.

        :param executable: The executable program.
        :param res_mgr: Optionally provide a results manager to save pass information.
        :param met_mgr: Optionally provide a metric manager to save pass information.
        :returns: Execution results.
        """

        if res_mgr is None:
            res_mgr = ResultManager()
        if met_mgr is None:
            met_mgr = MetricsManager()

        self.aggregator.clear()
        with self._hold_connection():
            for program in executable.programs:
                playback = self.engine.execute(program, met_mgr=met_mgr, **kwargs)
                self.aggregator.append(playback, executable.acquires)

        acquisitions = self.aggregator.finalise()

        return self.results_pipeline.run(
            acquisitions, res_mgr, met_mgr, package=executable, **kwargs
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

from qat.core.pass_base import PassManager
from qat.model.hardware_model import PhysicalHardwareModel as PydHardwareModel
from qat.model.target_data import TargetData
from qat.purr.compiler.hardware_models import QuantumHardwareModel
from qat.runtime.passes.analysis import IndexMappingAnalysis as PydIndexMappingAnalysis
from qat.runtime.passes.purr.analysis import IndexMappingAnalysis
from qat.runtime.passes.transform import (
    AcquisitionPostprocessing,
    AssignResultsTransform,
    ErrorMitigation,
    InlineResultsProcessingTransform,
    ResultTransform,
)


def get_default_results_pipeline(model: QuantumHardwareModel) -> PassManager:
    """Factory for creating the default results pipeline.

    :param model: The quantum hardware model to use for the pipeline.
    """

    return (
        PassManager()
        | AcquisitionPostprocessing()
        | InlineResultsProcessingTransform()
        | AssignResultsTransform()
        | ResultTransform()
        | IndexMappingAnalysis(model)
        | ErrorMitigation(model)
    )


def get_results_pipeline(model: PydHardwareModel, target_data: TargetData) -> PassManager:
    """Factory for creating the results pipeline.

    :param model: The quantum hardware model.
    """

    return (
        PassManager()
        | AcquisitionPostprocessing(target_data)
        | InlineResultsProcessingTransform()
        | AssignResultsTransform()
        | ResultTransform()
        | PydIndexMappingAnalysis(model)
        | ErrorMitigation(model)
    )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
from functools import singledispatchmethod

from qat.ir.gates.base import GateBase
from qat.ir.gates.native import X_pi_2, Z_phase, ZX_pi_4
from qat.ir.gates.operation import Barrier, Measure, Reset
from qat.ir.instructions import PhaseShift, QuantumInstruction, Synchronize
from qat.ir.measure import Acquire
from qat.ir.waveforms import Pulse, SampledWaveform, Waveform
from qat.middleend.decompositions.base import DecompositionBase
from qat.middleend.decompositions.gates import (
    DefaultGateDecompositions,
    GateDecompositionBase,
)
from qat.model.hardware_model import PhysicalHardwareModel


class PulseDecompositionBase(DecompositionBase):
    """Base object for implementing decompositions of :class:`NativeGate` and other qubit
    operations such as :class:`Measure`, :class:`Reset` and :class:`Barrier`.

    This handles decompositions of gates / operations to pulse level instructions. Details
    of gate-to-gate decompositions should not be specified here, but can be injected as
    :attr:`gate_decompositions` to allow for decomposition of higher-level gates into
    pulses.
    """

    end_nodes = (QuantumInstruction,)

    def __init__(
        self,
        gate_decompositions: GateDecompositionBase = None,
    ):
        """
        :param gate_decompositions: The DAG for decomposing higher-level, non-native gates
            can be provided as an optional argument. Uses the
            :class:`DefaultGateDecompositions` by default.
        """

        if not gate_decompositions:
            gate_decompositions = DefaultGateDecompositions()
        self.gate_decompositions = gate_decompositions

    @singledispatchmethod
    def decompose_op(self, gate: GateBase, model: PhysicalHardwareModel):
        """Implements the definition of a decomposition of a gate.

        The definition does not have to be in terms of native gates, but decompositions
        must form a DAG. This will call :attr:`gate_decompositions` for higher-level gates.
        Decompositon details of native gates and other qubit operations should be
        implemented here.
        """
        return self.gate_decompositions.decompose(gate)


class DefaultPulseDecompositions(PulseDecompositionBase):
    """Implements the standard gate-to-pulse decompositions used in OQC hardware.

    Provides decomposition rules for the native gate set {:class:`Z_phase`,
    :class:`X_pi_2`, :class:`ZX_pi_4`}. Gates which are not part of
    the native set are first decomposed into native gates using
    :class:`DefaultGateDecompositions`, and then decomposed into pulses using the rules
    provided here.
    """

    end_nodes = (QuantumInstruction,)

    def __init__(self, gate_decompositions: DecompositionBase = None):
        """
        :param gate_decompositions: The DAG for decomposing higher-level, non-native gates
            can be provided as an optional argument. Uses the:class:`DefaultGateDecompositions`
            by default.
        """

        if not gate_decompositions:
            gate_decompositions = DefaultGateDecompositions()
        self.gate_decompositions = gate_decompositions

    @singledispatchmethod
    def decompose_op(self, gate: GateBase, *args):
        """Implements the definition of a decomposition of a gate.

        The definition does not have to be in terms of native gates, but decompositions
        must form a DAG. This will call :attr:`gate_decompositions` for higher-level gates.
        Decompositon details of native gates and other qubit operations should be
        implemented here.
        """
        return self.gate_decompositions.decompose(gate)

    @decompose_op.register(X_pi_2)
    def _(self, gate: X_pi_2, model: PhysicalHardwareModel):
        """Decomposes the :class:`X_pi_2` intruction into a drive pulse."""

        qubit = model.qubit_with_index(gate.qubit)
        pulse_channel = qubit.drive_pulse_channel
        pulse_info = pulse_channel.pulse.model_dump()
        pulse_waveform = pulse_channel.pulse.waveform_type(**pulse_info)
        return [Pulse(targets=pulse_channel.uuid, waveform=pulse_waveform)]

    @decompose_op.register(Z_phase)
    def _(self, gate: Z_phase, model: PhysicalHardwareModel):
        """Decomposes the :class:`Z_phase` intruction into a :class:`PhaseShift` on all
        appropiate channels."""

        qubit = model.qubit_with_index(gate.qubit)
        pulse_channel = qubit.drive_pulse_channel
        instructions = [PhaseShift(targets=pulse_channel.uuid, phase=gate.theta)]
        for (
            qid,
            crc_pulse_channel,
        ) in qubit.cross_resonance_cancellation_pulse_channels.items():
            coupled_qubit = model.qubit_with_index(qid)
            cr_pulse_channel = coupled_qubit.cross_resonance_pulse_channels[gate.qubit]

            instructions.append(PhaseShift(targets=cr_pulse_channel.uuid, phase=gate.theta))
            instructions.append(
                PhaseShift(targets=crc_pulse_channel.uuid, phase=gate.theta)
            )
        return instructions

    @decompose_op.register(ZX_pi_4)
    def _(self, gate: ZX_pi_4, model: PhysicalHardwareModel):
        r"""Decomposes a :math:`ZX(\pi/4)` into cross-resonance pulses."""

        qubit1 = model.qubits[gate.qubit1]
        qubit2 = model.qubits[gate.qubit2]
        target1_pulse_channel = qubit1.cross_resonance_pulse_channels[gate.qubit2]
        target2_pulse_channel = qubit2.cross_resonance_cancellation_pulse_channels[
            gate.qubit1
        ]

        if target1_pulse_channel is None or target2_pulse_channel is None:
            raise ValueError(
                f"Tried to perform cross resonance on {str(qubit2)} that isn't linked to "
                f"{str(qubit1)}."
            )

        pulse_info = target1_pulse_channel.zx_pi_4_pulse.model_dump()
        if pulse_info is None:
            raise ValueError(
                f"No `zx_pi_4_pulse` available on {qubit1} with index {gate.qubit1}."
            )

        return [
            Synchronize(targets=[target1_pulse_channel.uuid, target2_pulse_channel.uuid]),
            Pulse(
                targets=target1_pulse_channel.uuid,
                waveform=target1_pulse_channel.zx_pi_4_pulse.waveform_type(**pulse_info),
            ),
            Pulse(
                targets=target2_pulse_channel.uuid,
                waveform=target1_pulse_channel.zx_pi_4_pulse.waveform_type(**pulse_info),
            ),
        ]

    @decompose_op.register(Barrier)
    def _(self, gate: Barrier, model: PhysicalHardwareModel):
        r"""Decomposes a :class:`Barrier` into a :class:`Synchronize` across pulse channels."""

        pulse_channel_ids = set()
        for qid in gate.qubits:
            target = model.qubit_with_index(qid)
            pulse_channel_ids.add(target.acquire_pulse_channel.uuid)
            pulse_channel_ids.add(target.measure_pulse_channel.uuid)
            qubit_pulse_channel_ids = [
                pulse_channel.uuid for pulse_channel in target.all_pulse_channels
            ]
            pulse_channel_ids.add(*qubit_pulse_channel_ids)
        return [Synchronize(targets=pulse_channel_ids)]

    @decompose_op.register(Measure)
    def _(self, gate: Measure, model: PhysicalHardwareModel):
        r"""Decomposes a :class:`Measure` object into a measure pulse and an acquisition."""
        # TODO: deal with complexities of the output_variable: classical registers, random
        # generation of names, etc... (COMPILER-285)
        # TODO: post processing responsibilities? should we have a "return mode" in the
        # measure instruction. The post-processing path is decided by acquire mode + return
        # mode. (COMPILER-287)

        qubit = model.qubit_with_index(gate.qubit)

        # Measure-related info.
        measure_channel = qubit.measure_pulse_channel
        measure_instruction = Pulse(
            targets=measure_channel.uuid,
            waveform=Waveform(**measure_channel.pulse.model_dump()),
        )

        # Acquire-related info.
        acquire_channel = qubit.acquire_pulse_channel
        acquire_duration = (
            measure_channel.pulse.width
            if acquire_channel.acquire.sync
            else acquire_channel.acquire.width
        ) - acquire_channel.acquire.delay
        if acquire_channel.acquire.use_weights is False:
            filter = Pulse(
                waveform=SampledWaveform(samples=acquire_channel.acquire.weights),
                duration=acquire_duration,
            )
        else:
            filter = None

        acquire_instruction = Acquire(
            targets=acquire_channel.uuid,
            duration=acquire_duration,
            mode=gate.mode,
            delay=acquire_channel.acquire.delay,
            filter=filter,
            output_variable=None,
        )

        return [
            Barrier(qubits=(gate.qubit)),
            measure_instruction,
            acquire_instruction,
        ]

    @decompose_op.register(Reset)
    def _(self, gate: Reset, model: PhysicalHardwareModel):
        r"""Decomposes a :class:`Reset` into a "passive reset" (i.e. just wait for the state
        to decohere).

        Not implemented for default pulse decompositions.
        """

        return NotImplementedError("The Reset operation is not yet implemented.")



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024-2025 Oxford Quantum Circuits Ltd

import numpy as np
from compiler_config.config import (
    CompilerConfig,
    ErrorMitigationConfig,
    InlineResultsProcessing,
    ResultsFormatting,
)

from qat.backend.qblox.acquisition import Acquisition
from qat.core.pass_base import TransformPass
from qat.core.result_base import ResultManager
from qat.executables import Executable
from qat.ir.instructions import Variable
from qat.model.hardware_model import PhysicalHardwareModel
from qat.model.target_data import TargetData
from qat.purr.backends.utilities import software_post_process_linear_map_complex_to_real
from qat.purr.compiler.error_mitigation.readout_mitigation import get_readout_mitigation
from qat.purr.compiler.hardware_models import QuantumHardwareModel
from qat.purr.compiler.instructions import AcquireMode, PostProcessType
from qat.runtime.passes.purr.analysis import IndexMappingResult
from qat.runtime.post_processing import apply_post_processing, get_axis_map
from qat.runtime.results_processing import binary_average, binary_count, numpy_array_to_list


class AcquisitionPostprocessing(TransformPass):
    """Uses the post-processing instructions from the executable package to process the
    results from the engine.

    The target machine will return the results in a format that depends on the specified
    :class:`AcquireMode`. However, it is often the case results need to be returned in an
    explicit format, e.g., as discriminated bits. To achieve this, extra software
    post-processing is needed.

    The post-processing that appears here is the same as the post-processing
    responsibilities taken on by the :class:`QuantumExecutionEngine` in
    :mod:`qat.purr.compiler.execution`.
    """

    def __init__(self, target_data: TargetData | None = None):
        """
        :param target_data: The target data is needed to know the sample time of the
            acquired results.
        """
        self.sample_time = (
            target_data.RESONATOR_DATA.sample_time if target_data is not None else None
        )

    def run(self, acquisitions: dict[str, any], *args, package: Executable, **kwargs):
        """
        :param acquisitions: The dictionary of results acquired from the target machine.
        :param package: The executable program containing the results-processing
            information should be passed as a keyword argument.
        """

        for output_variable, acquire in package.acquires.items():
            response = acquisitions[output_variable]

            # Starting from all axes, we iterate through each post-processing, keeping track
            # of what axes remain as we go
            response_axes = get_axis_map(acquire.mode, response)
            for pp in acquire.post_processing:
                response, response_axes = apply_post_processing(response, pp, response_axes)
            acquisitions[output_variable] = response

        return acquisitions


class QBloxAcquisitionPostProcessing(TransformPass):
    def run(
        self,
        playback: dict[str, dict[str, Acquisition]],
        *args,
        package: Executable,
        **kwargs,
    ):
        """
        Now that the combined playback is ready, we can compute and process results as required
        by customers. This requires loop nest information as well as post-processing and array shaping
        requirements.
        """

        results = {}
        for pulse_channel_id, acquisitions in playback.items():
            # TODO - Support multiple acquires target (unicity by name only is not wise)
            acquires = package.acquires
            for name, acquisition in acquisitions.items():
                scope_data = acquisition.acquisition.scope
                integ_data = acquisition.acquisition.bins.integration
                thrld_data = acquisition.acquisition.bins.threshold

                # TODO - COMPILER-860 - Safer and more flexible acquisition addressing
                acquire_data = acquires[name]
                if acquire_data.mode in [AcquireMode.SCOPE, AcquireMode.RAW]:
                    response = scope_data.path0.data + 1j * scope_data.path1.data
                elif acquire_data.mode == AcquireMode.INTEGRATOR:
                    response = integ_data.path0 + 1j * integ_data.path1
                else:
                    raise ValueError(f"Unrecognised acquire mode {acquire_data.mode}")

                post_procs, axes = acquire_data.post_processing, {}
                for pp in post_procs:
                    if pp.process_type == PostProcessType.LINEAR_MAP_COMPLEX_TO_REAL:
                        response, _ = software_post_process_linear_map_complex_to_real(
                            pp.args, response, axes
                        )
                    elif pp.process_type == PostProcessType.DISCRIMINATE:
                        # f : {0, 1} ----> {-1, 1}
                        #       x    |---> 1 - 2x
                        response = 1 - 2 * thrld_data

                response = response.reshape(acquire_data.shape)

                # TODO - COMPILER-860 - Safer and more flexible acquisition addressing
                if name in results:
                    raise ValueError(f"Key {name} already exists")
                results[name] = response

        return results


class InlineResultsProcessingTransform(TransformPass):
    """Uses :class:`InlineResultsProcessing` instructions from the executable package to
    format the acquired results in the desired format.
    """

    def run(self, acquisitions: dict[str, any], *args, package: Executable, **kwargs):
        """
        :param acquisitions: The dictionary of results acquired from the target machine.
        :param package: The executable program containing the results-processing
            information should be passed as a keyword argument.
        """
        # TODO: clean up imported utility

        for output_variable in acquisitions:
            target_values = acquisitions[output_variable]

            rp = package.acquires[output_variable].results_processing
            if rp is None:
                rp = InlineResultsProcessing.Experiment

            if InlineResultsProcessing.Raw in rp and InlineResultsProcessing.Binary in rp:
                raise ValueError(
                    "Raw and Binary processing attempted to be applied to "
                    f"{output_variable}. Only one should be selected."
                )

            # Strip numpy arrays if we're set to do so.
            if InlineResultsProcessing.NumpyArrays not in rp:
                target_values = numpy_array_to_list(target_values)

            # Transform to various formats if required.
            if InlineResultsProcessing.Binary in rp:
                target_values = binary_average(target_values)

            acquisitions[output_variable] = target_values

        return acquisitions


class AssignResultsTransform(TransformPass):
    """Processes :class:`Assign` instructions.

    As assigns are classical instructions they are not processed as a part of the quantum
    execution (right now). Read through the results dictionary and perform the assigns
    directly, return the results.

    Extracted from :meth:`purr.compiler.execution.QuantumExecutionEngine._process_assigns`.
    """

    def run(self, acquisitions: dict[str, any], *args, package: Executable, **kwargs):
        """
        :param acquisitions: The dictionary of results acquired from the target machine.
        :param package: The executable program containing the results-processing
            information should be passed as a keyword argument.
        """

        def recurse_arrays(results_map, value):
            """Recurse through assignment lists and fetch values in sequence."""
            if isinstance(value, list):
                return [recurse_arrays(results_map, val) for val in value]
            elif isinstance(value, tuple):
                return results_map[value[0]][value[1]]
            elif isinstance(value, str):
                return results_map[value]
            elif isinstance(value, Variable):
                return results_map[value.name]
            else:
                return value

        assigns = dict(acquisitions)
        for assign in package.assigns:
            assigns[assign.name] = recurse_arrays(assigns, assign.value)
        return {key: assigns[key] for key in package.returns}


class ResultTransform(TransformPass):
    """Transform the raw results into the format that we've been asked to provide. Look at
    individual transformation documentation for descriptions on what they do.

    Extracted from :meth:`qat.purr.compiler.runtime.QuantumRuntime._transform_results`.
    """

    def run(
        self,
        acquisitions: dict[str, any],
        *args,
        compiler_config: CompilerConfig,
        package: Executable,
        **kwargs,
    ):
        """
        :param acquisitions: The dictionary of results acquired from the target machine.
        :param compiler_config: The compiler config is needed to know how to process the
            results, and should be provided as a keyword argument.
        """
        # TODO: should the needed information in the compiler config be processed (in a
        # compilation pass) so that its not here. My opinion is that the compiler config
        # shouldn't make it to runtime.

        format_flags = (
            compiler_config.results_format or ResultsFormatting.DynamicStructureReturn
        )
        repeats = package.shots or TargetData.default().default_shots

        if len(acquisitions) == 0:
            return []

        def simplify_results(simplify_target):
            """To facilitate backwards compatability and being able to run low-level
            experiments alongside quantum programs we make some assumptions based upon
            form of the results.

            If all results have default variable names then the user didn't care about
            value assignment or this was a low-level experiment - in both cases, it
            means we can throw away the names and simply return the results in the order
            they were defined in the instructions.

            If we only have one result after this, just return that list directly
            instead, as it's probably just a single experiment.
            """
            if all([self._is_generated_name(k) for k in simplify_target.keys()]):
                if len(simplify_target) == 1:
                    return list(simplify_target.values())[0]
                else:
                    squashed_results = list(simplify_target.values())
                    if all(isinstance(val, np.ndarray) for val in squashed_results):
                        return np.array(squashed_results)
                    return squashed_results
            else:
                return simplify_target

        if ResultsFormatting.BinaryCount in format_flags:
            acquisitions = {
                key: binary_count(val, repeats) for key, val in acquisitions.items()
            }

        def squash_binary(value):
            if isinstance(value, int):
                return str(value)
            elif all(isinstance(val, int) for val in value):
                return "".join([str(val) for val in value])

        if ResultsFormatting.SquashBinaryResultArrays in format_flags:
            acquisitions = {key: squash_binary(val) for key, val in acquisitions.items()}

        # Dynamic structure return is an ease-of-use flag to strip things that you know
        # your use-case won't use, such as variable names and nested lists.
        if ResultsFormatting.DynamicStructureReturn in format_flags:
            acquisitions = simplify_results(acquisitions)
        return acquisitions

    @staticmethod
    def _is_generated_name(name: str) -> bool:
        """Check if a name is a generated name."""
        return name.startswith("generated_name_")


class ErrorMitigation(TransformPass):
    """Applies readout error mitigation to the results.

    Extracted from :meth:`qat.purr.compiler.runtime.QuantumRuntime._apply_error_mitigation`.
    """

    def __init__(self, hardware_model: QuantumHardwareModel | PhysicalHardwareModel):
        """
        :param hardware_model: The hardware model contains the error mitigation properties.
        """
        self.hardware_model = hardware_model

    def run(
        self,
        acquisitions: dict[str, any],
        res_mgr: ResultManager,
        *args,
        compiler_config: CompilerConfig,
        **kwargs,
    ):
        """
        :param acquisitions: The dictionary of results acquired from the target machine.
        :param res_mgr: The results manager is needed to look up the qubit-to-variable
            mapping.
        :param compiler_config: The compiler config is needed to know how to apply error
            mitigaiton, and should be provided as a keyword argument.
        """
        error_mitigation = compiler_config.error_mitigation

        if error_mitigation is None or error_mitigation == ErrorMitigationConfig.Empty:
            return acquisitions

        mapping = res_mgr.lookup_by_type(IndexMappingResult).mapping

        # TODO: add support for multiple registers
        # TODO: reconsider results length
        if len(acquisitions) > 1:
            raise ValueError(
                "Cannot have multiple registers in conjunction with readout error "
                "mitigation."
            )

        for mitigator in get_readout_mitigation(error_mitigation):
            new_acquisitions = mitigator.apply_error_mitigation(
                acquisitions, mapping, self.hardware_model
            )
            acquisitions[mitigator.name] = new_acquisitions
        return acquisitions



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

from compiler_config.config import CompilerConfig

from qat.core.metrics_base import MetricsManager
from qat.core.result_base import ResultManager
from qat.frontend.base import BaseFrontend
from qat.ir.instruction_builder import InstructionBuilder
from qat.purr.utils.logger import get_default_logger

logger = get_default_logger()


class QatFrontend(BaseFrontend):
    """A fallthrough frontend for accepting QAT IR. Rejects anything that isn't an
    :class:`InstructionBuilder`.

    :param model: The hardware model to use for compilation.
    """

    def __init__(self, model):
        self.model = model

    def check_and_return_source(self, src):
        """Checks if the source is a valid purr instruction builder.

        :param src: The source program, or path to the program.
        :returns: If the program is determined to not be valid, False is returned.
            Otherwise, the program is returned (and loaded if required).
        """
        if not isinstance(src, InstructionBuilder):
            return False

        if self.model and src.hw and self.model != src.hw:
            logger.info("InstructionBuilder model does not match the frontend model.")
            return False

        return src

    def emit(
        self,
        src: InstructionBuilder,
        res_mgr: ResultManager | None = None,
        met_mgr: MetricsManager | None = None,
        compiler_config: CompilerConfig | None = None,
        **kwargs,
    ) -> InstructionBuilder:
        """Passes through the instruction builder without modifying it."""

        src = self.check_and_return_source(src)
        if src is False:
            raise ValueError("Source is not a valid InstructionBuilder.")
        return src



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
from contextlib import contextmanager
from contextvars import ContextVar

from qat.core.config.session import QatSessionConfig
from qat.purr.qatconfig import QatConfig

_global_config: QatSessionConfig | None = None
_session_config: ContextVar[QatSessionConfig | None] = ContextVar(
    "session_config", default=None
)


def get_qatconfig() -> QatConfig:
    """Returns the global QatConfig"""
    global _global_config
    if _global_config is None:
        _global_config = QatConfig()
    return _global_config


def get_config() -> QatSessionConfig | QatConfig:
    """Returns the session config if set or the global QatConfig if not"""
    if (session_config := _session_config.get()) is not None:
        return session_config

    return get_qatconfig()


@contextmanager
def override_config(config: QatSessionConfig):
    token = _session_config.set(config)
    try:
        yield
    finally:
        _session_config.reset(token)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024-2025 Oxford Quantum Circuits Ltd
import re
from dataclasses import dataclass, field

from compiler_config.config import CompilerConfig

from qat.core.pass_base import AnalysisPass
from qat.core.result_base import ResultInfoMixin, ResultManager
from qat.executables import Executable
from qat.purr.backends.calibrations.remote import find_calibration
from qat.purr.compiler.builders import InstructionBuilder
from qat.purr.compiler.hardware_models import QuantumHardwareModel
from qat.purr.compiler.instructions import Acquire
from qat.purr.compiler.runtime import CalibrationWithArgs
from qat.runtime.passes.analysis import IndexMappingResult


@dataclass
class CalibrationAnalysisResult(ResultInfoMixin):
    calibration_executables: list[CalibrationWithArgs] = field(default_factory=list)


class CalibrationAnalysis(AnalysisPass):
    def run(
        self,
        ir: InstructionBuilder,
        res_mgr: ResultManager,
        *args,
        compiler_config: CompilerConfig,
        **kwargs,
    ):
        cal_blocks = [find_calibration(arg) for arg in compiler_config.active_calibrations]
        res_mgr.add(CalibrationAnalysisResult(cal_blocks))
        return ir


class IndexMappingAnalysis(AnalysisPass):
    """Determines a mapping from classical bit registers to qubits.

    Searches through the acquires of the package and determines their associated qubit.
    Also looks for classical registers of the form :code:`<clreg_name>[<clreg_index>]`.

    Supports both :class:`Executable` packages and :class:`InstructionBuilder`.
    """

    def __init__(self, model: QuantumHardwareModel):
        """:param model: The hardware model is needed for the qubit mapping."""
        self.model = model

    def run(
        self,
        acquisitions: dict[str, any],
        res_mgr: ResultManager,
        *args,
        package: InstructionBuilder | Executable,
        **kwargs,
    ):
        """
        :param acquisitions: The dictionary of results acquired from the target machine.
        :param res_mgr: The results manager to save the mapping.
        :param package: The executable program containing the results-processing
            information should be passed as a keyword argument.
        """
        # Determine a mapping from output variable - > qubit index.
        if isinstance(package, Executable):
            var_to_channel_map = self.var_to_physical_channel_executable(package)
        else:
            var_to_channel_map = self.var_to_physical_channel_qat_ir(package)
        var_to_qubit_map = self.var_to_qubit_map(var_to_channel_map)

        # Search for classical registers defined in e.g. QASM
        pattern = re.compile(r"(.*)\[([0-9]+)\]")
        for var in list(var_to_qubit_map.keys()):
            result = pattern.match(var)
            if result:
                var_to_qubit_map[result.group(2)] = var_to_qubit_map.pop(var)

        res_mgr.add(IndexMappingResult(mapping=var_to_qubit_map))
        return acquisitions

    @staticmethod
    def var_to_physical_channel_executable(package: Executable) -> dict[str, str]:
        return {var: acquire.physical_channel for var, acquire in package.acquires.items()}

    @staticmethod
    def var_to_physical_channel_qat_ir(package: InstructionBuilder) -> dict[str, str]:
        mapping: dict[str, str] = {}
        for instruction in package.instructions:
            if not isinstance(instruction, Acquire):
                continue
            mapping[instruction.output_variable] = instruction.channel.physical_channel_id
        return mapping

    def var_to_qubit_map(self, mapping: dict[str, str]):
        chan_to_qubit_map: dict[str, int] = {}
        for chan in set(mapping.values()):
            qubits = [
                qubit
                for qubit in self.model.qubits
                if qubit.get_acquire_channel().physical_channel_id == chan
            ]

            if len(qubits) == 0:
                raise ValueError(f"Could not find any qubits with acquire channel {chan}.")
            chan_to_qubit_map[chan] = qubits[0].index

        var_to_qubit_map = {var: chan_to_qubit_map[chan] for var, chan in mapping.items()}
        return var_to_qubit_map



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
import os

from piny import StrictMatcher


class VeryStrictMatcher(StrictMatcher):
    """
    Expand an environment variable of form ${VAR} with its value

    If value is not found, raises a ValueError
    """

    @staticmethod
    def constructor(loader, node):
        match = StrictMatcher.matcher.match(node.value)
        key = match.groups()[0]
        if key not in os.environ:
            raise ValueError(f"Environment Variable {key} not found")
        return os.environ[key]  # type: ignore



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
from qat.model.target_data import TargetData
from qat.pipelines.legacy.base import results_pipeline
from qat.pipelines.pipeline import ExecutePipeline
from qat.pipelines.updateable import PipelineConfig, UpdateablePipeline
from qat.purr.backends.echo import EchoEngine
from qat.purr.compiler.hardware_models import QuantumHardwareModel
from qat.purr.utils.logger import get_default_logger
from qat.runtime.legacy import LegacyRuntime

log = get_default_logger()


class LegacyEchoExecutePipeline(UpdateablePipeline):
    """A pipeline that executes compiled using the legacy echo backend.

    .. warning::

        This pipeline is for execution purposes only and does not compile programs. Please
        select an appropriate compilation pipeline if you wish to compile programs before
        execution.
    """

    @staticmethod
    def _build_pipeline(
        config: PipelineConfig,
        model: QuantumHardwareModel,
        target_data: TargetData | None = None,
        engine=None,
    ) -> ExecutePipeline:
        if engine is not None:
            log.warning(
                "The engine for the LegacyEchoExecutePipeline is expected to be provided by "
                "the model, and the provided engine will be ignored."
            )

        target_data = target_data if target_data is not None else TargetData.default()
        return ExecutePipeline(
            name=config.name,
            model=model,
            target_data=target_data,
            runtime=LegacyRuntime(
                engine=EchoEngine(model=model),
                results_pipeline=results_pipeline(model),
            ),
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

from qat.frontend.auto import (
    AutoFrontend as AutoFrontend,
    AutoFrontendWithFlattenedIR as AutoFrontendWithFlattenedIR,
)
from qat.frontend.base import BaseFrontend as BaseFrontend
from qat.frontend.custom import CustomFrontend as CustomFrontend
from qat.frontend.fallthrough import FallthroughFrontend as FallthroughFrontend
from qat.frontend.qasm import Qasm2Frontend as Qasm2Frontend, Qasm3Frontend as Qasm3Frontend
from qat.frontend.qat_ir import QatFrontend as QatFrontend
from qat.frontend.qir import QIRFrontend as QIRFrontend

DefaultFrontend = AutoFrontend
